﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Diagnostics;
using System.IO;

namespace Common
{
    public abstract class MossLayer
    {
        public abstract void ExportDefinitions(ExportAction exportAction);

        public abstract bool ImportScopeDisplayGroup(
            TScopeDisplayGroup groupElement, 
            string searchCollectionUrl,
            ProgramOptions.ConflictBehaviorType conflictBehavior);

        public abstract bool ImportScope(
            TScope scopeElement, string entityIdentifier, 
            bool isSiteScope, ProgramOptions.ConflictBehaviorType conflictBehavior,
            out int createdRules);

        public abstract bool ImportKeyword(
            TKeyword keywordElement, string searchCollectionUrl, 
            ProgramOptions.ConflictBehaviorType conflictBehavior, out int createdBestbets, out int createdSynonyms);

        public abstract bool ImportSearchSettings(
            TSearchSettings element, string siteCollectionUrl,
            ProgramOptions.ConflictBehaviorType conflictBehavior);

        public abstract bool ImportManagedProperties(TManagedProperty managedPropertyElement,
            string entityIdentifier,
            ProgramOptions.ConflictBehaviorType conflictBehavior,
            out int createdProperties, out int updatedProperties);


        
    }

    /// <summary>
    /// Defines the basic interface of an actionType, i.e an actionType which the tool can perform. This basic interface consists of:
    ///  - Keeping the program options
    ///  - A method PerformAction, which the main program calls to perform an actionType.
    ///  - A Factory method that creates an actionType based on a set of program option
    /// </summary>
    public abstract class ProgramAction
    {
        public ProgramAction(ProgramOptions options)
        {
            this.options = options;
        }

        abstract public void PerformAction();

        public ProgramOptions Options
        {
            get
            {
                return options;
            }
        }

        public static ProgramAction BuildAction(ProgramOptions options, MossLayer layer)
        {
            switch (options.ActionToPerform)
            {
                case ProgramOptions.ActionType.Export:
                    return new ExportAction(options, layer);
                case ProgramOptions.ActionType.Import:
                    return new ImportAction(options, layer);
                case ProgramOptions.ActionType.SaveSample:
                case ProgramOptions.ActionType.SaveSchema:
                    return new SaveInformationFileAction(options);
                default:
                    return null;
            }
        }

        private ProgramOptions options;
    }

    /// <summary>
    /// Besides the simple interface of Action, has a layer, which is a way to communicate with a specific version of MOSS. 
    /// Also all the common logic to import/export actions is here.
    /// </summary>
    public abstract class ComplexAction : ProgramAction
    {
        public ComplexAction(ProgramOptions options, MossLayer layer):base(options)
        {
            mossLayer = layer;
        }

        override public void PerformAction()
        {
            //Prints to the console and logs a summary with counts of all objects processed
            foreach (MossItem item in Enum.GetValues(typeof(MossItem)))
            {
                string itemString = item.ToString();
                if(Options.FastSearch && item == MossItem.ScopeRule)
                {
                    itemString = "ScopeFilter";
                }
                string message = String.Format("{0} {1} {2}s", ActionVerb, GetItemCount(item), itemString);
                Console.WriteLine(message);
                Logger.Log(TraceEventType.Information, message);
            }
        }

        abstract protected string ActionVerb
        {get;}

        public bool ShouldProcessSiteCollection(string url)
        {
            if (!String.IsNullOrEmpty(Options.SiteCollectionFilter))
            {
                return System.Text.RegularExpressions.Regex.Match(
                    url, Options.SiteCollectionFilter,
                    System.Text.RegularExpressions.RegexOptions.IgnoreCase).Success;
            }
            else //if we dont have a site filter we will process all sites, except if theres a search app filter
            {
                return String.IsNullOrEmpty(Options.SearchAppFilter);
            }
        }

        public bool ShouldProcessSearchApplication(string name)
        {
            if (!String.IsNullOrEmpty(Options.SearchAppFilter))
            {
                return System.Text.RegularExpressions.Regex.Match(
                    name, Options.SearchAppFilter,
                    System.Text.RegularExpressions.RegexOptions.IgnoreCase).Success;
            }
            else
            {
                return String.IsNullOrEmpty(Options.SiteCollectionFilter);
            }
        }

        public enum MossItem
        {
            Keyword,
            Bestbet,
            Synonym,
            Scope,
            ScopeRule,
            ScopeDisplayGroup,
            SearchSettings,
            ManagedProperty,
            UpdatedManagedProperty
        }

        private Dictionary<MossItem, int> itemCounter = new Dictionary<MossItem, int>();

        protected void IncreaseItemCount(MossItem item, int value)
        {
            int currentCount = 0;
            if (itemCounter.ContainsKey(item))
            {
                currentCount = itemCounter[item];
            }
            itemCounter[item] = currentCount + value;
        }
        protected void IncreaseItemCount(MossItem item)
        {
            IncreaseItemCount(item, 1);
        }

        public int GetItemCount(MossItem item)
        {
            int currentCount = 0;
            if (itemCounter.ContainsKey(item))
            {
                currentCount = itemCounter[item];
            }
            return currentCount;
        }

        protected MossLayer mossLayer;
    }

    public class NodeNames
    {
        public const string Prefix = "mss";
        public const string Namespace = "http://schemas.microsoft.com/SharePoint/Search/2010/datamigration";
        public const string SearchObjects = "SearchObjects";
        public const string SiteCollection = "SiteCollection";
        public const string SearchApplication = "SearchApplication";
        public const string Keyword = "Keyword";
        public const string Scope = "Scope";
        public const string ScopeDisplayGroup = "ScopeDisplayGroup";
        public const string Bestbet = "Bestbet";
        public const string Synonym = "Synonym";
        public const string AllContentScopeRule = "AllContentScopeRule";
        public const string UrlScopeRule = "UrlScopeRule";
        public const string PropertyQueryScopeRule = "PropertyQueryScopeRule";
        public const string SearchSettings = "SearchSettings";
        public const string ManagedProperty = "ManagedProperty";
    }

    public class ImportAction : ComplexAction
    {
        public ImportAction(ProgramOptions options, MossLayer layer)
            : base(options, layer)
        {
        }

        protected override string ActionVerb
        { get { return "Imported";} }

        public override void PerformAction()
        { 
            if ((Options.ActionToPerform & ProgramOptions.ActionType.Import) == 0)
            {
                return; //no importing do be done
            }

            string validationErrorMessage = XMLValidator.Validate(File.OpenText(Options.FileName).ReadToEnd());
            if (!String.IsNullOrEmpty(validationErrorMessage))
            {
                throw new ActionException(
                    "Cannot import definitions since the input file is invalid. The following syntax errors were found:" +
                    Environment.NewLine + validationErrorMessage);
            }

            TSearchObjects searchObjectsElement = DeserializeInputXml();
            if (searchObjectsElement.SearchApplication != null)
            {
                foreach (TSearchApplication searchApp in searchObjectsElement.SearchApplication)
                {
                    if (ShouldProcessSearchApplication(searchApp.Name))
                    {
                        ImportSearchAppicationInformation(searchApp);
                    }
                }
            }
            if (searchObjectsElement.SiteCollection != null)
            {
                foreach (TSiteCollection siteCollectionElement in searchObjectsElement.SiteCollection)
                {
                    if (ShouldProcessSiteCollection(siteCollectionElement.Url))
                    {
                        ImportSiteCollectionInformation(siteCollectionElement);
                    }
                }
            }
            base.PerformAction();
        }

        private TSearchObjects DeserializeInputXml()
        {
            FileStream fs = new FileStream(Options.FileName, FileMode.Open, FileAccess.Read);

            System.Xml.Serialization.XmlSerializer serializer = new System.Xml.Serialization.XmlSerializer(typeof(TSearchObjects));
            XmlReader reader = new XmlTextReader(fs);

            if (serializer.CanDeserialize(reader))
            {
                return (TSearchObjects)serializer.Deserialize(reader);
            }
            else
            {
                return null;
            }
        }

        private void ImportSiteCollectionInformation(TSiteCollection siteCollectionElement)
        {
            try
            {
                if ((Options.Entities & ProgramOptions.ActionEntities.Bestbet) != 0 &&
                    null != siteCollectionElement.Keyword)
                {
                    ProgressIndicator keywordProgress =
                           new ProgressIndicator(
                               string.Format("Importing keywords for Site [{0}]", siteCollectionElement.Url),
                               siteCollectionElement.Keyword.Count());
                    foreach (TKeyword keywordElement in siteCollectionElement.Keyword)
                    {
                        int createdBestbets = 0, createdSynonyms = 0;
                        if (mossLayer.ImportKeyword(
                                keywordElement, siteCollectionElement.Url, 
                                Options.ConflictBehavior, out createdBestbets, out createdSynonyms))
                        {
                            IncreaseItemCount(MossItem.Keyword);
                            IncreaseItemCount(MossItem.Bestbet, createdBestbets);
                            IncreaseItemCount(MossItem.Synonym, createdSynonyms);
                        }
                        keywordProgress.IncreaseCounter();
                    }
                }
                if ((Options.Entities & ProgramOptions.ActionEntities.Scope) != 0)
                {
                    Uri siteCollectionUri = new Uri(siteCollectionElement.Url);

                    if (null != siteCollectionElement.Scope)
                    {
                        ProgressIndicator scopeProgress =
                               new ProgressIndicator(
                                   string.Format("Importing scopes for [{0}]", siteCollectionElement.Url),
                                   siteCollectionElement.Scope.Count());

                        foreach (TScope scopeElement in siteCollectionElement.Scope)
                        {
                            int createdRules = 0;
                            if (mossLayer.ImportScope(
                                scopeElement, siteCollectionElement.Url, true,
                                Options.ConflictBehavior, out createdRules))
                            {
                                IncreaseItemCount(MossItem.Scope);
                                IncreaseItemCount(MossItem.ScopeRule, createdRules);
                            }
                            scopeProgress.IncreaseCounter();
                        }
                    }
                    if (null != siteCollectionElement.ScopeDisplayGroup)
                    {
                        ProgressIndicator groupProgress =
                              new ProgressIndicator(
                                  string.Format("Importing display groups for Site [{0}]", siteCollectionElement.Url),
                                  siteCollectionElement.ScopeDisplayGroup.Count());

                        foreach (TScopeDisplayGroup groupElement in siteCollectionElement.ScopeDisplayGroup)
                        {
                            if (mossLayer.ImportScopeDisplayGroup(groupElement, siteCollectionElement.Url, Options.ConflictBehavior))
                            {
                                IncreaseItemCount(MossItem.ScopeDisplayGroup);
                            }
                            groupProgress.IncreaseCounter();
                        }
                    }
                }
                if ((Options.Entities & ProgramOptions.ActionEntities.SearchSettings) != 0)
                {
                    Uri siteCollectionUri = new Uri(siteCollectionElement.Url);

                    if (null != siteCollectionElement.SearchSettings)
                    {
                        ProgressIndicator searchSettingsProgress =
                               new ProgressIndicator(
                                   string.Format("Importing search settings for [{0}]", siteCollectionElement.Url),
                                   1);
                        if (mossLayer.ImportSearchSettings(
                            siteCollectionElement.SearchSettings, siteCollectionElement.Url,
                            Options.ConflictBehavior))
                        {
                            IncreaseItemCount(MossItem.SearchSettings);
                        }
                        searchSettingsProgress.IncreaseCounter();
                    }
                }
            }
            catch (Exception ex)
            {   //if we cant get a site collection continue with the next one.
                Logger.Log(TraceEventType.Error, ex,
                    "Cannot import data for Site [{0}].",
                    siteCollectionElement.Url);
            }
        }   

        private void ImportSearchAppicationInformation(TSearchApplication searchApp)
        {
            try
            {
                if ((Options.Entities & ProgramOptions.ActionEntities.Scope) != 0 &&
                    null != searchApp.Scope)
                {
                    ProgressIndicator scopesProgress =
                           new ProgressIndicator(
                               string.Format("Importing scopes for Search App [{0}]", searchApp.Name),
                               searchApp.Scope.Count());
                    foreach (TScope scopeElement in searchApp.Scope)
                    {
                        int createdRules = 0;
                        if (mossLayer.ImportScope(scopeElement, searchApp.Name, false, Options.ConflictBehavior, out createdRules))
                        {
                            IncreaseItemCount(MossItem.Scope);
                            IncreaseItemCount(MossItem.ScopeRule, createdRules);
                        }
                        scopesProgress.IncreaseCounter();
                    }
                }
            }
            catch (Exception ex)
            {   //if we cant get a site collection continue with the next one.
                Logger.Log(TraceEventType.Error,ex,"Cannot import data for Search App [{0}].",searchApp.Name);
            }

            // Managed properties
            try {
                if ((Options.Entities & ProgramOptions.ActionEntities.ManagedProperties) != 0) {
                    ProgressIndicator scopesProgress =
                           new ProgressIndicator(
                               string.Format("Importing MPs for Search App [{0}]", searchApp.Name),
                               searchApp.ManagedProperty.Count());
                    foreach (TManagedProperty propertyElement in searchApp.ManagedProperty) {
                        int createdProperties = 0;
                        int updatedProperties = 0;
                        if (mossLayer.ImportManagedProperties(propertyElement, searchApp.Name, Options.ConflictBehavior, out createdProperties, out updatedProperties)) {
                            IncreaseItemCount(MossItem.ManagedProperty, createdProperties);
                            IncreaseItemCount(MossItem.UpdatedManagedProperty, updatedProperties);
                        }
                        scopesProgress.IncreaseCounter();
                    }
                }
            }
            catch (Exception ex) {   //if we cant get a site collection continue with the next one.
                Logger.Log(TraceEventType.Error, ex, "Cannot import data for Search App [{0}].", searchApp.Name);
            }

        }
        
    }

    public class ExportAction : ComplexAction
    {
        public ExportAction(ProgramOptions options, MossLayer layer)
            : base(options, layer)
        {
            xmlDocument = new XmlDocument();
            XmlDeclaration xmlDeclaration = xmlDocument.CreateXmlDeclaration("1.0", "utf-8", null); 
            // Create the root element
            rootNode = xmlDocument.CreateNode(XmlNodeType.Element,
                NodeNames.Prefix + ":" + NodeNames.SearchObjects,
                NodeNames.Namespace);
            xmlDocument.InsertBefore(xmlDeclaration, xmlDocument.DocumentElement);
            xmlDocument.AppendChild(rootNode);
            AddChild(rootNode, "FileVersion", VersionInformation.FileVersion, XmlNodeType.Attribute);
            Debug.Assert(null != rootNode);
        }
        protected override string ActionVerb
        { get { return "Exported";} }

        public override void PerformAction()
        {
            ExportDefinitions();
            Debug.Assert(XMLValidator.Validate(xmlDocument.OuterXml) == String.Empty);
            Logger.Log(TraceEventType.Verbose, XMLValidator.Validate(xmlDocument.OuterXml));
            xmlDocument.Save(Options.FileName);
            base.PerformAction();
        }

        private void ExportDefinitions()
        {
            if ((Options.ActionToPerform & ProgramOptions.ActionType.Export) == 0)
            {
                return; //no exporting do be done
            }

            mossLayer.ExportDefinitions(this);
        }

        public void WriteKeyword(string siteCollectionUrl, TKeyword keyword)
        {
            XmlNode siteCollectionNode = GetSiteCollectionNode(siteCollectionUrl);
            XmlNode keywordNode = AddChild(siteCollectionNode, NodeNames.Keyword, "", XmlNodeType.Element);

            AddChild(keywordNode, "Term", keyword.Term, XmlNodeType.Element);
            AddChild(keywordNode, "Contact", keyword.Contact, XmlNodeType.Element);
            AddChild(keywordNode, "StartDate", DateToString(keyword.StartDate), XmlNodeType.Element);
            AddChild(keywordNode, "EndDate", DateToString(keyword.EndDate), XmlNodeType.Element);
            AddChild(keywordNode, "ReviewDate", DateToString(keyword.ReviewDate), XmlNodeType.Element);
            AddChild(keywordNode, "Definition", keyword.Definition, XmlNodeType.Element);

            if (keyword.Synonym != null)
            {
                foreach (string synonym in keyword.Synonym)
                {
                    AddChild(keywordNode, NodeNames.Synonym, synonym, XmlNodeType.Element);
                    IncreaseItemCount(MossItem.Synonym);
                }
            }
            if (keyword.Bestbet != null)
            {
                foreach (TBestbet bestbet in keyword.Bestbet)
                {
                    XmlNode bbNode = AddChild(keywordNode, NodeNames.Bestbet, "", XmlNodeType.Element);
                    AddChild(bbNode, "Title", bestbet.Title, XmlNodeType.Element);
                    AddChild(bbNode, "Url", bestbet.Url, XmlNodeType.Element);
                    AddChild(bbNode, "Description", bestbet.Description, XmlNodeType.Element);
                    IncreaseItemCount(MossItem.Bestbet);
                }
            }
            siteCollectionNode.AppendChild(keywordNode);
            IncreaseItemCount(MossItem.Keyword);
        }

        public void WriteSearchSettings(string siteCollectionUrl, string searchCenterUrl, string dropdownMode, string searchResultsUrl)
        {
            XmlNode siteCollectionNode = GetSiteCollectionNode(siteCollectionUrl);
            XmlNode searchSettingsNode = AddChild(siteCollectionNode, NodeNames.SearchSettings, "", XmlNodeType.Element);
            bool addnode = false;
            if (!string.IsNullOrEmpty(searchCenterUrl))
            {
                addnode = true;
                AddChild(searchSettingsNode, "SearchCenterUrl", searchCenterUrl, XmlNodeType.Element);
            }
            if (!string.IsNullOrEmpty(dropdownMode))
            {
                addnode = true;
                AddChild(searchSettingsNode, "DropdownMode", dropdownMode, XmlNodeType.Element);
            }
            if (!string.IsNullOrEmpty(searchResultsUrl))
            {
                addnode = true;
                AddChild(searchSettingsNode, "SearchResultsUrl", searchResultsUrl, XmlNodeType.Element);
            }


            if (addnode)
            {
                siteCollectionNode.AppendChild(searchSettingsNode);
                IncreaseItemCount(MossItem.SearchSettings);
            }
        }

        public void WriteScope(string entityIdentifier, bool isSiteCollection, TScope scope)
        {
            XmlNode node;
            if (isSiteCollection)
            {
                node = GetSiteCollectionNode(entityIdentifier);
            }
            else
            {
                node = GetSearchServiceNode(entityIdentifier);
            }
            WriteScopeToNode(scope, node);
        }

        private void WriteScopeToNode(TScope scope, XmlNode parentNode)
        {
            XmlNode scopeNode = AddChild(parentNode, NodeNames.Scope, "", XmlNodeType.Element);
            IncreaseItemCount(MossItem.Scope);

            AddChild(scopeNode, "Name", scope.Name, XmlNodeType.Element);
            AddChild(scopeNode, "Description", scope.Description, XmlNodeType.Element);
            AddChild(scopeNode, "AlternateResultsPage", scope.AlternateResultsPage, XmlNodeType.Element);
            AddChild(scopeNode, "CompilationType", scope.CompilationType.ToString(), XmlNodeType.Element);
            AddChild(scopeNode, "Filter", scope.Filter, XmlNodeType.Element);
            AddChild(scopeNode, "ScopeKey", scope.ScopeKey.ToString(), XmlNodeType.Element);
            if (scope.AllContentScopeRule != null)
            {
                foreach (TAllContentScopeRule rule in scope.AllContentScopeRule)
                {
                    AddChild(scopeNode, NodeNames.AllContentScopeRule, "", XmlNodeType.Element);
                    IncreaseItemCount(MossItem.ScopeRule);
                }
            }
            if (scope.UrlScopeRule != null)
            {
                foreach (TUrlScopeRule rule in scope.UrlScopeRule)
                {
                    XmlNode ruleNode = AddChild(scopeNode, NodeNames.UrlScopeRule, "", XmlNodeType.Element);
                    IncreaseItemCount(MossItem.ScopeRule);
                    AddChild(ruleNode, "FilterBehavior", rule.FilterBehavior.ToString(), XmlNodeType.Element);
                    AddChild(ruleNode, "UrlScopeRuleType", rule.UrlScopeRuleType.ToString(), XmlNodeType.Element);
                    AddChild(ruleNode, "MatchingString", rule.MatchingString, XmlNodeType.Element);
                }
            }
            if (scope.PropertyQueryScopeRule != null)
            {
                foreach (TPropertyQueryScopeRule rule in scope.PropertyQueryScopeRule)
                {
                    XmlNode ruleNode = AddChild(scopeNode, NodeNames.PropertyQueryScopeRule, "", XmlNodeType.Element);
                    IncreaseItemCount(MossItem.ScopeRule);
                    AddChild(ruleNode, "FilterBehavior", rule.FilterBehavior.ToString(), XmlNodeType.Element);
                    AddChild(ruleNode, "Property", rule.Property, XmlNodeType.Element);
                    AddChild(ruleNode, "Value", rule.Value, XmlNodeType.Element);
                }
            }
        }

        public void WriteManagedProperty(string entityIdentifier, TManagedProperty property) {
            XmlNode node = GetSearchServiceNode(entityIdentifier);
            IncreaseItemCount(MossItem.ManagedProperty);
            WriteManagedPropertyToNode(property, node);
        }

        private void WriteManagedPropertyToNode(TManagedProperty property, XmlNode parentNode) {
            XmlNode scopeNode = AddChild(parentNode, NodeNames.ManagedProperty, "", XmlNodeType.Element);
            

            AddChild(scopeNode, "Name", property.Name, XmlNodeType.Element);
            AddChild(scopeNode, "Description", property.Description, XmlNodeType.Element);
            AddChild(scopeNode, "Type", property.Type.ToString(), XmlNodeType.Element);
            AddChild(scopeNode, "EnabledForScoping", property.EnabledForScoping.ToString().ToLower(), XmlNodeType.Element);
            AddChild(scopeNode, "HasMultipleValues", property.HasMultipleValues.ToString().ToLower(), XmlNodeType.Element);
            AddChild(scopeNode, "FullTextQueriable", property.FullTextQueriable.ToString().ToLower(), XmlNodeType.Element);
            foreach (var item in property.CrawledProperty) {
                XmlElement e = AddChild(scopeNode, "CrawledProperty", string.Empty, XmlNodeType.Element) as XmlElement;
                e.SetAttribute("Name", item.Name);
                e.SetAttribute("Propset", item.Propset);
                e.SetAttribute("VariantType", item.VariantType.ToString());
                e.SetAttribute("Order", item.Order.ToString());
            }
        }
        public void WriteScopeDisplayGroup(string siteCollectionUrl, TScopeDisplayGroup group)
        {
            //ignores Out of Box Display Groups
            if (IsBuiltInAndUnmodified(group))
            {
                return;
            }

            XmlNode siteCollectionNode = GetSiteCollectionNode(siteCollectionUrl);
            XmlNode groupNode = AddChild(siteCollectionNode, NodeNames.ScopeDisplayGroup, "", XmlNodeType.Element);
            IncreaseItemCount(MossItem.ScopeDisplayGroup);
            AddChild(groupNode, "Name", group.Name, XmlNodeType.Element);
            AddChild(groupNode, "Description", group.Description, XmlNodeType.Element);

            if (group.ScopeKey != null)
            {
                foreach (string scope in group.ScopeKey)
                {
                    AddChild(groupNode, "ScopeKey", scope, XmlNodeType.Element);
                }
            }
            AddChild(groupNode, "DefaultScopeKey", group.DefaultScopeKey, XmlNodeType.Element);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="group"></param>
        /// <returns>
        /// True if the displaygroup is exactely like an one that ships with MOSS 
        /// ("Search Dropdown" or "Advanced Search"). Note: This will only work on 
        /// english site collections. For other languages it will return false, even 
        /// if the group is out of the box.
        /// </returns>
        private static bool IsBuiltInAndUnmodified(TScopeDisplayGroup group)
        {
            bool isOOBGroup = false;

            if (group != null && !string.IsNullOrEmpty(group.Name))
            {
                if (group.Name.Equals("Search Dropdown", StringComparison.OrdinalIgnoreCase))
                {
                    if (group.ScopeKey != null && group.ScopeKey.Length == 2)
                    {
                        if (String.Equals(group.ScopeKey[0], Util.BuiltInScopeKeys.AllSites, StringComparison.OrdinalIgnoreCase) &&
                            String.Equals(group.ScopeKey[1], Util.BuiltInScopeKeys.People, StringComparison.OrdinalIgnoreCase) &&
                            String.Equals(group.DefaultScopeKey, Util.BuiltInScopeKeys.AllSites, StringComparison.OrdinalIgnoreCase))
                        {
                            isOOBGroup = true;
                        }
                    }
                }
                else if (group.Name.Equals("Advanced Search", StringComparison.OrdinalIgnoreCase))
                {
                    if (group.ScopeKey != null && group.ScopeKey.Length == 1)
                    {
                        if (String.Equals(group.ScopeKey[0], Util.BuiltInScopeKeys.AllSites, StringComparison.OrdinalIgnoreCase) &&
                            String.Equals(group.DefaultScopeKey, Util.BuiltInScopeKeys.AllSites, StringComparison.OrdinalIgnoreCase))
                        {
                            isOOBGroup = true;
                        }
                    }
                }
            }
            return isOOBGroup;
        }

        private XmlNode GetSiteCollectionNode(string siteCollectionUrl)
        {
            XmlNode siteCollectionNode = null;
            siteCollectionNodes.TryGetValue(siteCollectionUrl, out siteCollectionNode);
            if (null == siteCollectionNode)
            {
                siteCollectionNode = AddChild(rootNode, "SiteCollection", "", XmlNodeType.Element);
                AddChild(siteCollectionNode, "Url", siteCollectionUrl, XmlNodeType.Attribute);
                rootNode.AppendChild(siteCollectionNode);
                siteCollectionNodes[siteCollectionUrl] = siteCollectionNode;
            }

            return siteCollectionNode;
        }

        private XmlNode GetSearchServiceNode(string searchService)
        {
            XmlNode searchServicenNode = null;
            searchServiceNodes.TryGetValue(searchService, out searchServicenNode);
            if (null == searchServicenNode)
            {
                searchServicenNode = AddChild(rootNode, NodeNames.SearchApplication, "", XmlNodeType.Element);
                AddChild(searchServicenNode, "Name", searchService, XmlNodeType.Attribute);
                rootNode.AppendChild(searchServicenNode);
                searchServiceNodes[searchService] = searchServicenNode;
            }

            return searchServicenNode;
        }

        private XmlNode AddChild(XmlNode parentNode, string nodeName, string nodeValue, XmlNodeType nodeType)
        {
            switch (nodeType)
            {
                case XmlNodeType.Attribute:
                    XmlAttribute newAttribute = xmlDocument.CreateAttribute(nodeName);
                    newAttribute.InnerText = nodeValue;
                    parentNode.Attributes.Append(newAttribute);
                    return newAttribute;
                case XmlNodeType.Element:
                    XmlNode newElement = xmlDocument.CreateElement(nodeName);
                    //makes sure we don't create empty nodes like <tag></tag>, but rather <tag />
                    if (!String.IsNullOrEmpty(nodeValue))
                    {
                        newElement.InnerText = nodeValue;
                    }
                    parentNode.AppendChild(newElement);
                    return newElement;
                default:
                    return null;
            }
        }

        private static string DateToString(DateTime date)
        {
            return date.ToString("s");
        }

        private Dictionary<string, XmlNode> siteCollectionNodes = new Dictionary<string, XmlNode>();
        private Dictionary<string, XmlNode> searchServiceNodes = new Dictionary<string, XmlNode>();

        private XmlDocument xmlDocument;
        private XmlNode rootNode;
    }

    public class SaveInformationFileAction : ProgramAction
    {
        public SaveInformationFileAction(ProgramOptions options)
            : base(options)
        {
            if (options.ActionToPerform == ProgramOptions.ActionType.SaveSample)
            {
                resourceName = "XMLFileSample.xml";
            }
            else if (options.ActionToPerform == ProgramOptions.ActionType.SaveSchema)
            {
                resourceName = "XmlInputFileSchema.xsd";
            }
        }

        public override void PerformAction()
        {
            Debug.Assert(null != resourceName);
            string resourceContents = Util.GetResourceString(resourceName);
            if (null != resourceContents)
            {
                if (Options.ActionToPerform == ProgramOptions.ActionType.SaveSample)
                {
                    Debug.Assert(XMLValidator.Validate(resourceContents) == String.Empty);
                }
                File.WriteAllText(Options.FileName, resourceContents);
            }
        }

        string resourceName;

    }

    /// <summary>
    /// Holds the program command line switches. It is initialized with args form Main
    /// </summary>
    public class ProgramOptions
    {
        [Flags]
        public enum ActionType
        {
            None = 0x00,
            Import = 0x01,
            Export = 0x02,
            SaveSample = 0x04,
            SaveSchema = 0x08
        }

        [Flags]
        public enum ActionEntities
        {
            None = 0x00,
            Scope = 0x01,
            Bestbet = 0x02,
            SearchSettings = 0x04,
            ManagedProperties = 0x08,
            All = Scope + Bestbet + SearchSettings + ManagedProperties
        }

        public enum ConflictBehaviorType
        {
            Continue = 0,
            Overwrite,
            Prompt
        }

        public ProgramOptions(string[] args)
        {
            executableName = Path.GetFileNameWithoutExtension(
                Environment.GetCommandLineArgs()[0]);

            for (int i = 0, n = args.Length; i < n; i++)
            {
                string arg = args[i].ToLowerInvariant();
                switch (arg)
                {
                    case "-import":
                        SetActionType(ActionType.Import);
                        break;
                    case "-export":
                        SetActionType(ActionType.Export);
                        break;
                    case "-savesample":
                        SetActionType(ActionType.SaveSample);
                        break;
                    case "-saveschema":
                        SetActionType(ActionType.SaveSchema);
                        break;
                    case "-bestbet":
                        Entities |= ActionEntities.Bestbet;
                        break;
                    case "-scope":
                        Entities |= ActionEntities.Scope;
                        break;
                    case "-searchsettings":
                        Entities |= ActionEntities.SearchSettings;
                        break;
                    case "-managedproperty":
                        Entities |= ActionEntities.ManagedProperties;
                        break;
                    case "-all":
                        Entities |= ActionEntities.All;
                        break;
                    case "-fastsearch":
                        FastSearch = true;
                        break;
                    default:
                        //handles parameters of the form -parameter=value
                        if (arg.StartsWith("-sitefilter="))
                        {
                            SiteCollectionFilter = args[i].Substring("-sitefilter=".Length);
                            continue;
                        }
                        else if (arg.StartsWith("-searchappfilter="))
                        {
                            SearchAppFilter = args[i].Substring("-searchappfilter=".Length);
                            continue;
                        }
                        else if (arg.StartsWith("-conflictbehavior="))
                        {
                            ConflictBehavior = ParseBehavior(arg.Substring("-conflictbehavior=".Length));
                            continue;
                        }

                        //handles the filename or invalid parameters
                        else if (null == FileName && !arg.StartsWith("-"))
                        {
                            FileName = args[i];
                        }
                        else
                        {
                            ResetOptions();
                        }
                        return;
                }
            }
        }

        private void SetActionType(ActionType actionType)
        {
            if (actionType != ActionType.None)
            {
                ActionToPerform = actionType;
            }
            else
            {
                ResetOptions();//more than one action specified. That is invalid.
            }
 
        }

        private void ResetOptions()
        {
            ActionToPerform = ActionType.None;
            Entities = ActionEntities.None;
            SiteCollectionFilter = String.Empty;
            SearchAppFilter = String.Empty;
            FastSearch = false;
            this.fileName = null;
        }

        public string BrandString
        {
            get
            {
                return "Search Migration Tool v" + Common.VersionInformation.ShortVersion;
            }
        }

        public string GetUsageString()
        {
            return String.Format(BrandString + Environment.NewLine + Environment.NewLine +
@"Usage: 
{0} -action [-objectKind] [-filter=value] [-conflictBehavior=value] [-fastSearch] filename

---action must be one of:
  import     - Imports search objects.
  export     - Exports search objects.
  saveSample - Saves an example of an input/output xml file.
  saveSchema - Saves the xml schema (xsd) that represents the format of the
               input/output xml file.

---objectKind must be specified for the import or export actions. Use it to
specify which search objects should be handled in the operation. More than one
can be specified at the same time. It can have the following values:
  bestbet        - Processes keywords, bestbets and synonyms.
  scope          - Processes scopes, scope rules and scope display groups.
  searchsettings - Processes the search settings, including search center URL, 
                   dropdown mode and search results URL. This option is not
                   supported on Sharepoint Server 2007 and will be ignored
                   for that product.
  all            - Processes all of the above.

---filter is a way to specify filters for the site collections or search
applications to process. .NET regular expressions are used to do the matching
and it is case insentitive. You can specify both filters at the same time. 
value cannot have spaces, filter can be:
  siteFilter=value      - Only the site collections whose URL matches the
                          value will be processed.
  searchAppFilter=value - Only the searchService instances whose name matches
                          the value will be processed.
If only one filter is specified, the other type of entity will not be 
processed. E.g. -siteFilter=http://sharepoint will process no search 
applications.
Some filter examples:
  -siteFilter=http://sharepoint will match http://sharepoint/ or 
   http://sharepoint/sites/test, but not http://test/sharepoint
  -searchAppFilter=search.[12] will match ""Search 1"" but not ""Search 3""

---conflictBehavior is an optional way to specify how the tool should handle
existing objects that conflict with objects being imported in an import
operation. It can have one of the follwing values:
  conflictBehavior=continue  - Continue is the default. When a conflict
                               arises, an error is logged,the existing object
                                is untouched, and the operation continues
                               with the next one.
  conflictBehavior=overwrite - When a conflict arises, the existing object is
                               deleted, and a new one is created in its place.
  conflictBehavior=prompt    - When a conflict arises, the tool will prompt
                               with the options of continuing and overwriting.
                               This option may require extended monitoring on
                               a large farm.

---fastSearch will specificy that the action performed will be done using the
FAST Search Server for Sharepoint 2010 administration interface. This is only
supported with import action.

---Filename is a required parameter.  All actions have either an input or 
output file.

This tool supports the export action on both Microsoft Office SharePoint 
Server 2007 and Microsoft SharePoint Server 2010. The import action is only
supported for Microsoft SharePoint Server 2010.

Example Uses:
{0} -export -scope -siteFilter=http://sharepoint C:\filename.xml
{0} -export -scope -bestbet C:\filename.xml
{0} -import -bestbet -conflictBehavior=prompt C:\filename.xml
{0} -saveSample C:\sample.xml
{0} -saveSchema C:\schema.xsd",
            executableName);
        }

        public bool HasValidOptions()
        {
            return
                ActionToPerform != ActionType.None &&
                (Entities != ActionEntities.None || ActionToPerform == ActionType.SaveSample || ActionToPerform == ActionType.SaveSchema) &&
                FileName != null;
        }

        private string executableName;

        public ActionType ActionToPerform
        {
            get;
            private set;
        }
        public ActionEntities Entities
        {
            get;
            private set;
        }

        public bool FastSearch
        {
            get;
            private set;
        }

        private string fileName;
     
        public string FileName
        {
            get
            { 
                return this.fileName;
            }
            private set
            {
                bool fileNameValid = false;
                try
                {
                    new System.IO.FileInfo(value);
                    fileNameValid = true;
                }
                catch
                {
                }
                if (fileNameValid)
                {
                    this.fileName = value;
                }
                else
                {
                    ResetOptions();
                }
            }
        }
        public string SearchAppFilter
        {
            get;
            private set;
        }
        public string SiteCollectionFilter
        {
            get;
            private set;
        }

        public ConflictBehaviorType ConflictBehavior
        {
            get;
            private set;
        }

        private ConflictBehaviorType ParseBehavior(string argument)
        {
            switch (argument.ToLowerInvariant())
            {
                case "overwrite":
                    return ConflictBehaviorType.Overwrite;
                case "prompt":
                    return ConflictBehaviorType.Prompt;
                case "continue":
                default:
                    return ConflictBehaviorType.Continue;
            }
        }

        /// <summary>
        /// Handles a conflict based on the passed conflitc behavior. 
        /// Uses label for providing a prompt to the user if it is specified
        /// in conflictBehavior
        /// </summary>
        /// <param name="conflictBehavior"></param>
        /// <param name="label">The name of the thing that conflit is about. Ex: Keyword [house]</param>
        /// <returns>True the the tool should go ahead with the actionType and 
        /// overwite the existing object, and False if it should be ignored</returns>
        public static bool HandleConflict(ConflictBehaviorType conflictBehavior, string entityLabel)
        {
            switch(conflictBehavior)
            {
                default:
                case ConflictBehaviorType.Continue:
                    return false;
                case ConflictBehaviorType.Overwrite:
                    return true;
                case ConflictBehaviorType.Prompt:
                    Console.WriteLine();
                    Console.WriteLine(
                        "There is already an object {0}. Do you want to "+
                        "overwrite it? Responding no will skip it. Responding "+
                        "yes will delete the current object and create a new "+
                        "one in its place.", entityLabel);
                    return Util.PromptYesNo();
            }
        }

        /// <summary>
        /// returns a string describing the actions the program will take. This usefull for 
        /// prompting the user for confirmation before taking any actionType.
        /// </summary>
        /// <returns></returns>
        public string GetProgramActionDescription()
        {
            StringBuilder actions = new StringBuilder();
            actions.AppendLine(BrandString);
            actions.Append("Program Action: ");
            if ((ActionToPerform & ActionType.Import) != 0)
            {
                actions.AppendLine("Import");
            }
            else if ((ActionToPerform & ActionType.Export) != 0)
            {
                actions.AppendLine("Export");
            }
            else if ((ActionToPerform & ActionType.SaveSample) != 0)
            {
                actions.AppendLine("Save Sample Xml File");
            }
            else if ((ActionToPerform & ActionType.SaveSchema) != 0)
            {
                actions.AppendLine("Save Xml Schema for input/ouput file");
            }
            else
            {
                actions.AppendLine("None");
            }

            if (Entities != ActionEntities.None)
            {
                actions.Append("SharePoint objects considered:");
                if ((Entities & ActionEntities.Bestbet) != 0)
                {
                    actions.Append(" Bestbets");
                }
                if ((Entities & ActionEntities.Scope) != 0)
                {
                    actions.Append(" Scopes");
                }
                if ((Entities & ActionEntities.SearchSettings) != 0)
                {
                    actions.Append(" SearchSettings");
                }
                if ((Entities & ActionEntities.ManagedProperties) != 0) {
                    actions.Append(" ManagedProperties");
                }
                actions.AppendLine();
                if ((ActionToPerform & ActionType.Import) != 0)
                {
                    actions.Append("Conflict Behavior: ");
                    actions.AppendLine(ConflictBehavior.ToString());
                }
                if (!String.IsNullOrEmpty(SiteCollectionFilter))
                {
                    actions.Append("Site Collection Filter: ");
                    actions.AppendLine(SiteCollectionFilter);
                }
                if (!String.IsNullOrEmpty(SearchAppFilter))
                {
                    actions.Append("Search Application Filter: ");
                    actions.AppendLine(SearchAppFilter);
                }
                if(FastSearch)
                {
                    actions.AppendLine("FAST Search Processing Enabled");
                }
            }

            if (!String.IsNullOrEmpty(FileName))
            {
                actions.Append("Filename: ");
                actions.AppendLine(FileName);
            }
            return actions.ToString();
        }
    }

    /// <summary>
    /// This is the error thow by our code. We create a new class so that we don't have to 
    /// catch the general exception type
    /// </summary>
    public class ActionException : Exception
    {
        public ActionException(string message)
            : base(message)
        { }
    }
}