﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Reflection;
using System.Diagnostics;

namespace Common
{
    public class Util
    {
        public class BuiltInScopeKeys
        {
            public const string People = "0";
            public const string AllSites = "1";
            static public bool IsBuiltIn(string key)
            {
                return key == People || key == AllSites; 
            }
            static public int KeyToId(string key)
            {
                if (key == Util.BuiltInScopeKeys.People)
                {
                    return Util.BuiltInScopeIds.People;
                }
                else if (key == Util.BuiltInScopeKeys.AllSites)
                {
                    return Util.BuiltInScopeIds.AllSites;
                }
                else
                {
                    return -1;
                }
            }
        }
        public class BuiltInScopeIds
        {
            public const int People = 0;
            public const int AllSites = 1;
        }

        public static Stream GetResourceStream(string resourceId)
        {
            System.Reflection.Assembly assembly = System.Reflection.Assembly.GetAssembly(typeof(Util));
            if (assembly != null)
            {
                return assembly.GetManifestResourceStream(typeof(Util), resourceId);
            }
            return null;
        }

        public static string GetResourceString(string resourceId)
        {

            Stream stream = GetResourceStream(resourceId);
            if (null != stream)
            {
                return (new StreamReader(stream)).ReadToEnd();
            }
            else
            {
                return null;
            }
        }

        public static List<T> IEnumeratorToList<T>(System.Collections.IEnumerable collection)
        {
            List<T> list = new List<T>();
            if (collection != null)
            { 
                foreach (T item in collection)
                {
                    list.Add(item);
                }
            }
            return list;
        }

        /// <summary>
        /// Converts from one type of enum to another. Will only work if both enum 
        /// types have the same named items
        /// </summary>
        /// <typeparam name="TFrom"></typeparam>
        /// <typeparam name="TTo"></typeparam>
        /// <param name="enumValue"></param>
        /// <returns></returns>
        public static TTo MarshalEnum<TFrom, TTo>(TFrom enumValue) 
        {
            return (TTo)Enum.Parse(typeof(TTo), enumValue.ToString());
        }

        /// <summary>
        /// 
        /// </summary>
        /// <returns>true if the user selected yes.</returns>
        public static bool PromptYesNo()
        {
            Console.Write("[Y or N] then press [enter]:");

            while (true)
            {
                string readString = Console.ReadLine();
                if (String.Equals(readString, "Y", StringComparison.InvariantCultureIgnoreCase) ||
                    String.Equals(readString, "YES", StringComparison.InvariantCultureIgnoreCase))
                {
                    return true;
                }
                else if (String.Equals(readString, "N", StringComparison.InvariantCultureIgnoreCase) ||
                         String.Equals(readString, "NO", StringComparison.InvariantCultureIgnoreCase))
                {
                    return false;
                }
                Console.Write("[Y or N] then press [enter]:");
            }
        }
    }

    /// <summary>
    /// This class will track the progress of a operation with known finite number of steps.
    /// It will use tyhe console to print an increasing counter that tracks the operation
    /// </summary>
    public class ProgressIndicator
    {
        /// <summary>
        /// The progress counter looks like this:
        /// Performing Action: 3/10 (30%)
        /// In this example:
        ///     "Performing Action" is an label passed in construction
        ///     10 is the total number of steps (numberSteps parameter)
        ///     3 is the number of completed steps
        /// </summary>
        /// <param name="label"></param>
        /// <param name="numberSteps"></param>
        public ProgressIndicator(string label, int numberSteps)
        {
            if (numberSteps <= 0)
            {
                return;
            }
            this.label = label;
            this.numberSteps = numberSteps;
            WriteProgress();
        }

        /// <summary>
        /// Advance one step and print progress
        /// </summary>
        public void IncreaseCounter()
        {
            if (numberSteps <= 0)
            {
                return;
            }

            Debug.Assert(currentStep < numberSteps);
            currentStep++;
            WriteProgress();
            if (currentStep == numberSteps)
            {
                Console.WriteLine(" Done!");
            }
        }

        private string label;
        private int numberSteps;
        private int currentStep;
        private void WriteProgress()
        {
            if (numberSteps <= 0)
            {
                return;
            }

            Console.SetCursorPosition(0, Console.CursorTop);
            string progressNumbers = String.Format(": {0}/{1} ({2:0}%)",
                currentStep,
                numberSteps,
                (float)currentStep * 100.0 / (float)numberSteps);

            //make sure the whole progress strings fits one line. Does so by 
            //cutting the label and adding "...".
            //Also leaves one space empty at the right of the string to 
            //make sure that the cursor does not go to the next line
            string adjustedLabel;
            if (label.Length + progressNumbers.Length > Console.WindowWidth - 1)
            {
                adjustedLabel = label.Substring(
                    0,
                    Console.WindowWidth - progressNumbers.Length - 4) + "...";
            }
            else
            {
                adjustedLabel = label;
            }

            Console.Write("{0}{1}", adjustedLabel, progressNumbers);
        }
    }
    public class Logger
    {
        public static string FileName { get; set; }

        static public void Log(TraceEventType level, string message, params object[] args)
        {
            Write(level,message, args);
            theWriter.WriteLine();
        }

        static public void Log(TraceEventType level,Exception exception, string message, params object[] arg0)
        {
            Write(level,message, arg0);
            theWriter.WriteLine(" {0}",exception.ToString());
        }

        public static void Close()
        {
            theWriter.Flush();
            theWriter.Dispose();
        }
        
        public static int MessageCount(TraceEventType eventType)
        {
            if(messageCount.ContainsKey(eventType))
            {
                return messageCount[eventType];
            }
            else
            {
                messageCount.Add(eventType,0);
                return 0;
            }
        }
        public static int TotalMessageCount
        {
            get
            {
                return messageCount.Values.Sum();
            }
        }
        private static void IncreaseMessageCount(TraceEventType eventType)
        {
            if(messageCount.ContainsKey(eventType))
            {
                messageCount[eventType] = messageCount[eventType]+1;
            }
            else
            {
                messageCount.Add(eventType,1);
            }
        }
        static private void Write(TraceEventType level, string message, params object[] args)
        {
            if (theWriter == null)
            {
                Debug.Assert(!String.IsNullOrEmpty(FileName));
                FileInfo fileInfo = new FileInfo(FileName);
                theWriter = fileInfo.CreateText();
            }
            theWriter.Write("{0}\t{1}\t", //prints a header for each line
                DateTime.Now.ToString("hh:mm:ss"),
                level.ToString());
            theWriter.Write(message, args);
            IncreaseMessageCount(level);
        }
        private static StreamWriter theWriter;
        private static Dictionary<TraceEventType, int> messageCount = new Dictionary<TraceEventType, int>();
    }
}
