extern alias O14;

using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Collections.Specialized;
using Common;
using Microsoft.Office.Server.Search.Administration;
using Microsoft.Office.Server.Search.Query;
using Microsoft.SharePoint;
using Microsoft.SharePoint.Search.Extended.Administration.Keywords;
using Microsoft.SharePoint.Search.Extended.Administration.Common;
using Keyword=Microsoft.SharePoint.Search.Extended.Administration.Keywords.Keyword;
using O14Layer = O14::Layers.O14Layer;

namespace Layers
{
    /// <summary>
    /// This class handles import of keywords and scopes for Fast Search Server
    /// for SharePoint 2010.
    /// 
    /// Keywords are imported into FS14 the admin database exclusively using the
    /// FASTAdminProxy property of the SSA or SSA proxy. Scopes are imported
    /// into O14 using the scope import implemented in O14Layer, with additional
    /// generation of FQL filters for each scope implemented in this class.
    /// </summary>
    public class FS14Layer : O14Layer
    {
        private string currentSearchServiceApplicationName;
        private string currentSiteCollectionUrl;
        private SearchSettingGroup currentSearchSettingGroup;

        /// <summary>
        /// The maximum date for StartDate/EndDate properties that can be
        /// handled by the FAST keywords admin GUI.
        /// </summary>
        private static readonly DateTime FAST_MAX_DATE = new DateTime(8900, 12, 31);

        /// <summary>
        /// This contains the relationship between a Search App and the managed
        /// properties on the FAST back-end the Search App is configured with
        /// </summary>
        private Dictionary<String, StringCollection> _fastManagedProperties = new Dictionary<string, StringCollection>();

        /// <summary>
        /// Prepare a site collection for use by getting (or adding) the
        /// SearchSettingGroup for use by keyword handling, and managed
        /// properties for use by scope import.
        /// 
        /// If the site url is equal to the previous site url, no actions
        /// will be performed and the previous values will be reused.
        /// 
        /// On failure, the currentSearchSettingGroup instance field will
        /// be set to null and the error logged.
        /// </summary>
        /// <param name="searchCollectionUrl">site collection url</param>
        private void PrepareToUseSiteCollection(String searchCollectionUrl)
        {
            if (!string.Equals(currentSiteCollectionUrl, searchCollectionUrl, StringComparison.InvariantCultureIgnoreCase))
            {
                try
                {
                    // Connect to the FAST backend using the FASTAdminProxy property of a SSA proxy.
                    // If no backend is located, throw an exception
                    using (SPSite site = new SPSite(searchCollectionUrl))
                    {
                        String guid = site.ID.ToString();
                        var ssaProxy = GetSearchAppProxyForSite(site);
                        var adminProxy = ssaProxy.FASTAdminProxy;
                        if (adminProxy == null)
                        {
                            throw new InvalidOperationException("No FASTAdminProxy found for SSA");
                        }

                        // get or create a search setting group
                        var ssgs = adminProxy.KeywordContext.SearchSettingGroups;
                        currentSearchSettingGroup = ssgs.ContainsSearchSettingGroup(guid)
                                   ? ssgs.GetSearchSettingGroup(guid)
                                   : ssgs.AddSearchSettingGroup(guid);

                        // Try to fetch managed properties from FAST Search back-end, if search app is configured for it
                        currentSearchServiceApplicationName = ssaProxy.Name;
                        if (!_fastManagedProperties.ContainsKey(ssaProxy.Name))
                        {
                            _fastManagedProperties.Add(ssaProxy.Name, FetchManagedPropertiesFromFastBackend(ssaProxy));
                        }
                        else
                        {
                            Logger.Log(TraceEventType.Information,
                                "Using cached Managed Properties for Site [{0}] in Search App [{1}]'s FAST Search back-end. {2} managed properties cached.",
                                searchCollectionUrl, ssaProxy.Name, _fastManagedProperties[ssaProxy.Name].Count);
                        }
                    }
                }
                catch (Exception exception)
                {
                    currentSearchSettingGroup = null;
                    currentSearchServiceApplicationName = null;
                    Logger.Log(TraceEventType.Error, exception,
                        "Cannot import FAST Search information for Site [{0}] because I could not get a reference to it.",
                        searchCollectionUrl);
                }
                currentSiteCollectionUrl = searchCollectionUrl;
            }
        }

        /// <summary>
        /// Fetch all managed properties from the FAST back-end associated with
        /// this FASTAdminProxy. Results are cached according to the ssaName
        /// given.
        /// </summary>
        /// <param name="ssaProxy">The SSA Proxy to fetch FAST specific managed properties from</param>
        /// <returns>A StringCollection with all managed properties from the FAST back-end lowercased</returns>
        private static StringCollection FetchManagedPropertiesFromFastBackend(SearchServiceApplicationProxy ssaProxy)
        {
            StringCollection managedProperties = new StringCollection();

            if (ssaProxy != null)
            {
                try
                {
                    using (Query query = new KeywordQuery(ssaProxy))
                    {
                        query.ResultsProvider = SearchProvider.FASTSearch;
                        PropertyInformation[] queryProperties = query.GetProperties();
                        if (queryProperties != null)
                        {
                            Logger.Log(TraceEventType.Information,
                                       "Fetching Managed Properties from Search App [{0}]'s FAST Search back-end",
                                       ssaProxy.Name);
                            foreach (PropertyInformation pi in queryProperties)
                            {
                                managedProperties.Add(pi.Name.ToLowerInvariant());
                            }
                            Logger.Log(TraceEventType.Information,
                                       "Imported {0} Managed Properties from Search App [{1}]'s FAST Search back-end",
                                       managedProperties.Count, ssaProxy.Name);

                        }
                    }
                }
                catch (Exception e)
                {
                    Logger.Log(TraceEventType.Error, e,
                               "Unable to fetch managed properties from Search App [{0}]'s FAST Search back-end. Will continue without validation of managed properties.",
                               ssaProxy.Name);
                }
            }
            else
            {
                Logger.Log(TraceEventType.Verbose,
                       "FetchManagedPropertiesFromFastBackend: Search App is null");
            }

            return managedProperties;
        }

        /// <summary>
        /// Generates FQL Scope Filters for the given Search Service Application
        /// or site collection using FastScopeUtil.CreateFastScopeFilter().
        /// </summary>
        /// <param name="scopeElement">Element containing scope rules from xml</param>
        /// <param name="ssa">Search Service Application to fetch managed properties for</param>
        /// <param name="siteCollectionUrl">Site collection to fetch managed properties for</param>
        /// <returns>Generated FQL scope filter</returns>
        protected override string GenerateScopeFilter(TScope scopeElement, SearchServiceApplication ssa, string siteCollectionUrl)
        {
            string entity;
            StringCollection fetchedManagedProperties;
            SearchServiceApplicationProxy ssaProxy;
            if (ssa != null)
            {
                ssaProxy = GetSearchAppProxyByName(ssa.Name);
                entity = ssa.Name;
                
            }
            else
            {
                using (SPSite site = new SPSite(siteCollectionUrl))
                {
                    ssaProxy = GetSearchAppProxyForSite(site);
                    entity = ssaProxy.Name;
                }
            }

            // fetch managed properties for SSAs where this has not been done already (sites are handled in PrepareToUseSiteCollection)
            if (!_fastManagedProperties.ContainsKey(entity))
            {
                fetchedManagedProperties = FetchManagedPropertiesFromFastBackend(ssaProxy);
                _fastManagedProperties.Add(entity, fetchedManagedProperties);
                if (fetchedManagedProperties == null || fetchedManagedProperties.Count == 0)
                {
                    Logger.Log(TraceEventType.Warning, "FS14Layer::GenerateScopeFilter: FAST ManagedProperties is null/empty for entity [{0}]." +
                        "Filters will be created without verifying if the managed properties exist in the FAST back-end.",
                        entity);
                }
            }

            StringCollection managedProperties;
            // fetch managed properties from cache
            _fastManagedProperties.TryGetValue(entity, out managedProperties);

            // generate scope filter
            return FastScopeUtil.CreateFastScopeFilter(scopeElement, managedProperties, entity);
        }

        public override void ExportDefinitions(ExportAction exportAction)
        {
            throw new NotSupportedException("Exporting from Fast Search Server for SharePoint 2010 is  not supported!");
        }

        public override bool ImportKeyword(TKeyword keywordElement, string searchCollectionUrl, ProgramOptions.ConflictBehaviorType conflictBehavior, out int createdBestbets, out int createdSynonyms)
        {
            // Prepare the search setting group (fetch or create) and get the keywords collection
            createdBestbets = 0;
            createdSynonyms = 0;
            PrepareToUseSiteCollection(searchCollectionUrl);
            if (currentSearchSettingGroup == null)
            {
                return false;
            }
            var keywords = currentSearchSettingGroup.Keywords;

            // check if keyword exists, and prompt for overwrite (i.e. deletion)
            if (keywords.ContainsKeyword(keywordElement.Term))
            {
                //we have a conflict since the keyword already exists
                bool shouldOverwrite = ProgramOptions.HandleConflict(conflictBehavior, String.Format("Keyword [{0}]", keywordElement.Term));
                if (shouldOverwrite)
                {
                    try
                    {
                        keywords.RemoveKeyword(keywordElement.Term);
                        Logger.Log(TraceEventType.Information, "Deleted Keyword [{0}] for [{1}].", keywordElement.Term, searchCollectionUrl);
                    }
                    catch (Exception exception)
                    {
                        Logger.Log(TraceEventType.Error,exception,
                            "Cannot delete Keyword [{0}] for [{1}].",
                            keywordElement.Term,
                            searchCollectionUrl);
                        return false;
                    }
                }
                else
                {
                    Logger.Log(TraceEventType.Error,
                         "Skipped Keyword [{0}] in Site [{1}] in the input file because it already existed. " +
                         "You can change the treatment of such conflicts with option conflictBehavior",
                         keywordElement.Term, searchCollectionUrl);
                    return false;
                }
            }

            // create keyword
            Keyword keyword;
            try
            {
                keyword = keywords.AddKeyword(keywordElement.Term);
                keyword.Definition = keywordElement.Definition;
            }
            catch (Exception exception)
            {
                Logger.Log(TraceEventType.Error,exception,
                    "Cannot create Keyword [{0}] in Site [{1}].",
                    keywordElement.Term, searchCollectionUrl);
                return false;
            }

            // synonyms
            if (keywordElement.Synonym != null)
            {
                foreach (string synonymElement in keywordElement.Synonym)
                {
                    try
                    {
                        var syn = keyword.Synonyms.AddSynonym(synonymElement, SynonymExpansionType.TwoWay);
                        syn.StartDate = keywordElement.StartDate;
                        syn.EndDate = keywordElement.EndDate < FAST_MAX_DATE ? keywordElement.EndDate : FAST_MAX_DATE;
                        createdSynonyms++;
                    }
                    catch (AdminException exception)
                    {
                        Logger.Log(TraceEventType.Error,exception,
                           "Cannot create Synonym [{0}] for Keyword [{1}] in Site [{2}].",
                           synonymElement, keywordElement.Term, searchCollectionUrl);
                    }
                }
            }

            // best bets
            if (keywordElement.Bestbet != null)
            {
                foreach (TBestbet bestbetElement in keywordElement.Bestbet)
                {
                    try
                    {
                        var bestbet = keyword.BestBets.AddBestBet(bestbetElement.Title, new Uri(bestbetElement.Url));
                        bestbet.Position = (short) (createdBestbets + 1);
                        bestbet.Description = bestbetElement.Description;
                        bestbet.StartDate = keywordElement.StartDate;
                        bestbet.EndDate = keywordElement.EndDate < FAST_MAX_DATE ? keywordElement.EndDate : FAST_MAX_DATE;
                        createdBestbets++;
                    }
                    catch (AdminException exception)
                    {
                        Logger.Log(TraceEventType.Error, exception,
                           "Cannot create BestBet [{0}] for Keyword [{1}] in  Site [{2}].",
                           bestbetElement.Url, keywordElement.Term, searchCollectionUrl);
                    }
                }
            }

            // if we get here, the keyword is created (even if some synonym/best bet might have failed)
            return true;
        }
    }
}
