using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Diagnostics;
using System.Linq;
using System.Text;
using Common;

namespace Layers
{
    /// <summary>
    /// FAST specific utility class to build FQL filters from scope rules
    /// </summary>
    public class FastScopeUtil
    {
        private static readonly string KEYWORD_AND = "and";
        private static readonly string KEYWORD_OR = "or";
        private static readonly string KEYWORD_NOT = "not";
        private static readonly string KEYWORD_ANDNOT = "andnot";
        private static readonly string PATH_PROPERTY = "path";
        private static readonly string DOMAIN_PROPERTY = "domain";
        private static readonly string KEYWORD_STARTS_WITH = "starts-with";
        private static readonly string PROPERTY_SEPARATOR = ":";
        private static readonly string LogPrefix = "FastUtil::CreateFastScopeFilter:";
        internal static Int32 MaxCharactersInFilter = 2047;

        /// <summary>
        /// Create a FQL filter string from the scope rules in the TScope element.
        /// </summary>
        /// <param name="scopeElement">The TScope element holding the scope rules to generate the filter from</param>
        /// <param name="managedProperties">The managed properties found on the FAST back-end associated with the Search App</param>
        /// <param name="entity">The entity to create filters for; either a Search App or site collection uri</param>
        /// <returns>An FQL filter string based on the scope rules</returns>
        public static string CreateFastScopeFilter(TScope scopeElement, StringCollection managedProperties, string entity)
        {

            bool managedPropertiesSet = true;

            if (scopeElement == null)
                return "";

            // if there are no rules specified there can still be a filter, i.e a FAST only scope
            // in that case return the filter to preserve it
            if ((scopeElement.PropertyQueryScopeRule == null || scopeElement.PropertyQueryScopeRule.Length == 0) &&
                (scopeElement.UrlScopeRule == null || scopeElement.UrlScopeRule.Length == 0))
                return scopeElement.Filter;

            if (managedProperties == null || managedProperties.Count == 0)
            {
                managedPropertiesSet = false;
            }

            var requireList = new List<string>();
            var includeList = new List<string>();
            var excludeList = new List<string>();

            if (scopeElement.UrlScopeRule != null)
            {
                foreach (var rule in scopeElement.UrlScopeRule)
                {
                    string expression = "";
                    string property = "";
                    switch (rule.UrlScopeRuleType)
                    {
                        case TUrlScopeRuleType.Folder:
                            property = PATH_PROPERTY;
                            expression = PATH_PROPERTY + ":" + KEYWORD_STARTS_WITH + "(\"" +
                                         rule.MatchingString.ToLowerInvariant() + "\")";
                            break;
                        case TUrlScopeRuleType.HostName:
                            property = DOMAIN_PROPERTY;
                            expression = DOMAIN_PROPERTY + ":" + KEYWORD_STARTS_WITH + "(\"" +
                                         rule.MatchingString.ToLowerInvariant() + "\")";
                            break;
                        case TUrlScopeRuleType.Domain:
                            property = DOMAIN_PROPERTY;
                            expression = DOMAIN_PROPERTY + ":\"" + rule.MatchingString.ToLowerInvariant() + "\"";
                            break;
                    }

                    if (managedPropertiesSet && !managedProperties.Contains(property))
                    {
                        Logger.Log(TraceEventType.Error,
                                   "{0} Cannot create Filter for [{1}, {2}] to Scope [{3}] in {4}."
                                   + " Managed property [{5}] does not exist on FAST back-end.",
                                   LogPrefix, rule.UrlScopeRuleType.ToString(), rule.MatchingString,
                                   scopeElement.Name, entity, property);
                        return "";
                    }

                    switch (rule.FilterBehavior)
                    {
                        case TFilterBehavior.Exclude:
                            excludeList.Add(expression);
                            break;
                        case TFilterBehavior.Include:
                            includeList.Add(expression);
                            break;
                        case TFilterBehavior.Require:
                            requireList.Add(expression);
                            break;
                        default:
                            break;
                    }
                }
            }

            if (scopeElement.PropertyQueryScopeRule != null)
            {
                foreach (var rule in scopeElement.PropertyQueryScopeRule)
                {
                    if (managedPropertiesSet && !managedProperties.Contains(rule.Property.ToLowerInvariant()))
                    {
                        Logger.Log(TraceEventType.Error,
                                   "{0} Cannot add Property Rule [{1}={2}] to Scope [{3}] in {4}. Managed Property [{1}] does not exist on FAST back-end.",
                                   LogPrefix, rule.Property, rule.Value, scopeElement.Name, entity);
                        return "";
                    }
                    string expression = rule.Property.ToLowerInvariant() + PROPERTY_SEPARATOR + "\"" + rule.Value.ToLowerInvariant() + "\"";
                    switch (rule.FilterBehavior)
                    {
                        case TFilterBehavior.Exclude:
                            excludeList.Add(expression);
                            break;
                        case TFilterBehavior.Include:
                            includeList.Add(expression);
                            break;
                        case TFilterBehavior.Require:
                            requireList.Add(expression);
                            break;
                        default:
                            break;
                    }
                }
            }

            string fqlFilter = BuildFilter(scopeElement.Name, excludeList, includeList, requireList);

            if (fqlFilter.Length > MaxCharactersInFilter)
            {
                Logger.Log(TraceEventType.Error, "{0} The length of the generated filter is too long ({1}). Max length is {2}.",
                    LogPrefix, fqlFilter.Length, MaxCharactersInFilter);
                fqlFilter = "";
            }

            return fqlFilter;
        }

        internal static string BuildFilter(String scopeElementName, List<string> excludeList, List<string> includeList, List<string> requireList)
        {
            string fqlFilter = "";
            string orFilter = BuildFqlOrFilter(includeList);
            string andFilter = BuildFqlAndFilter(requireList);
            string excludeFilter = BuildFqlNotFilter(excludeList);

            if (!String.IsNullOrEmpty(orFilter) && !String.IsNullOrEmpty(andFilter))
            {
                // to support both include and require rules AND the orFilter and the andFilter
                // the result will be and(andFilter, orFilter)
                fqlFilter = CombineFilters(KEYWORD_AND, orFilter, andFilter);
            }
            else if (!String.IsNullOrEmpty(andFilter))
            {
                fqlFilter = andFilter;
            }
            else if (!String.IsNullOrEmpty(orFilter))
            {
                fqlFilter = orFilter;
            }
            else
            // no include or require filters exist, return just the excludelist
            {
                var notList = new List<string>();
                foreach (string excludeTerm in excludeList)
                {
                    notList.Add(String.Format("not({0})", excludeTerm));
                }
                return BuildFqlOrFilter(notList);
            }

            // include/require exists, as well as excludes. Combine them into an andnot-filter and return
            if (excludeList.Count > 0)
            {
                excludeList.Insert(0, fqlFilter);
                return BuildFqlAndNotFilter(excludeList);
            }

            // return just the include/require-filter
            return fqlFilter;
        }

        /// <summary>
        /// Combine two FQL-filters
        /// </summary>
        /// <param name="command">The command to cobine the filters with, e.g. 'and'</param>
        /// <param name="filter1">string representation of filter 1, e.g. or(apples, pears)</param>
        /// <param name="filter2">string representation of filter 2, e.g. and(bananas, oranges)</param>
        /// <returns>The filters combined, e.g. and(or(apples, pears), and(bananas, oranges))</returns>
        internal static String CombineFilters(string command, string filter1, string filter2)
        {
            if (String.IsNullOrEmpty(filter1))
                return filter2;
            if (String.IsNullOrEmpty(filter2))
                return filter1;

            var sb = new StringBuilder();
            sb.Append(command);
            sb.Append("(");
            sb.Append(filter1);
            sb.Append(", ");
            sb.Append(filter2);
            sb.Append(")");
            return sb.ToString();
        }

        /// <summary>
        /// Build a FQL and-filter
        /// </summary>
        /// <param name="queryElements">The operand(s) to build an AND over</param>
        /// <returns>A FQL AND-filter, e.g. and("apples", "pears")</returns>
        internal static String BuildFqlAndFilter(List<string> queryElements)
        {
            return BuildGenericFqlFilter(KEYWORD_AND, queryElements);
        }

        /// <summary>
        /// Build a FQL or-filter
        /// </summary>
        /// <param name="queryElements">The operand(s) to build an OR over</param>
        /// <returns>A FQL OR-filter, e.g. or("apples", "pears")</returns>
        internal static String BuildFqlOrFilter(List<string> queryElements)
        {
            return BuildGenericFqlFilter(KEYWORD_OR, queryElements);
        }

        /// <summary>
        /// Build a FQL not-filter
        /// </summary>
        /// <param name="queryElements">The operand(s) to not match</param>
        /// <returns>A FQL not-filter, e.g. not("apples", "pears")</returns>
        internal static String BuildFqlNotFilter(List<string> queryElements)
        {
            return BuildGenericFqlFilter(KEYWORD_NOT, queryElements);
        }

        /// <summary>
        /// Build an FQL andnot-filter.
        /// Only documents that match the first operand and NOT match the second operand will be returned.
        /// The format accepts more than 2 operands, where all operands but the first one are part of the NOT condition. 
        /// </summary>
        /// <param name="queryElements">List with the operand to match on index 0, and the operands
        /// to not match in the rest, e.g. {apples, pears, bananas}</param>
        /// <returns>andnot("apples", "pears", "bananas")</returns>
        internal static String BuildFqlAndNotFilter(List<string> queryElements)
        {
            // group the not-parameters
            return BuildGenericFqlFilter(KEYWORD_ANDNOT, queryElements);
        }

        /// <summary>
        /// Creates a valid query using a specific FQL command. For example,
        /// <i>genericQuery("and", {"microsoft", "corporation"})</i> returns the
        /// string <i>and("microsoft", "corporation")</i>.
        /// </summary>
        /// <param name="command">An FQL command (e.g. "and")</param>
        /// <param name="parameters">One or more parameters to the command (e.g. "microsoft","corporation")</param>
        /// <returns></returns>
        internal static string BuildGenericFqlFilter(string command, List<string> parameters)
        {
            if (command != KEYWORD_NOT && parameters != null && parameters.Count == 1)
                return parameters[0] ?? "";

            var sb = new StringBuilder();

            int tokens = 0;

            if (parameters != null && parameters.Count > 0 && command != null)
            {
                sb.Append(command);
                sb.Append("(");

                foreach (string p in parameters)
                {
                    if (!string.IsNullOrEmpty(p))
                    {
                        if (tokens > 0)
                        {
                            sb.Append(", ");
                        }

                        sb.Append(p);
                        tokens++;
                    }
                }
                sb.Append(")");
            }

            if (tokens > 0)
            {
                return sb.ToString();
            }
            // If all the parameters were empty strings or null, just return an empty string. 
            // Returning "command()" doesn't make sense, as it will give an syntax error in backend.
            return "";
        }
    }
}
