﻿using System;
using System.Collections.Generic;
using System.Diagnostics;

using Microsoft.SharePoint;
using Microsoft.SharePoint.Administration;
using Microsoft.Office.Server;
using Microsoft.Office.Server.Search.Administration;

using Common;

namespace Layers
{
    public class O12Layer : MossLayer
    {
        public override void ExportDefinitions(ExportAction exportAction)
        {
            if ((exportAction.Options.Entities & ProgramOptions.ActionEntities.Scope) != 0)
            {
                //TODO: As of now this code will retrieve all shared scopes, but will not split them by SSP
                //as it should have, so we just put them all together under one SearchApplication node.
                //In the future if possible we should somehow know which of those belong to which SSP, and
                //create a separate one for each of those.
                string searchAppName = "Search Service Application";
                if (exportAction.ShouldProcessSearchApplication(searchAppName))
                {
                    ServerContext serverContext = ServerContext.Default;
                    SearchContext searchContext = SearchContext.GetContext(serverContext);
                    ExportSharedScopesInformation(searchAppName, searchContext, exportAction);
                }
            }
            foreach (SPSite siteCollection in ListSiteCollections())
            {
                if(siteCollection == null)
                    continue;

                //ignores site collections that don't match the filter
                if (!exportAction.ShouldProcessSiteCollection(siteCollection.Url))
                {
                    continue;
                }

                if ((exportAction.Options.Entities & ProgramOptions.ActionEntities.Bestbet) != 0)
                {
                    ExportKeywords(siteCollection, exportAction);
                }

                if ((exportAction.Options.Entities & ProgramOptions.ActionEntities.Scope) != 0)
                {
                    ExportSiteScopesInformation(siteCollection, exportAction);
                }
            }
        }

        public override bool ImportScopeDisplayGroup(
            TScopeDisplayGroup groupElement, string searchCollectionUrl, 
            ProgramOptions.ConflictBehaviorType conflictBehavior)
        {
            throw new NotSupportedException("Importing into an O12 farm is unsupported!");
        }
        public override bool ImportScope(
            TScope scopeElement, string entityIdentifier, bool isSiteScope, 
            ProgramOptions.ConflictBehaviorType conflictBehavior, out int createdRules)
        {
            throw new NotSupportedException("Importing into an O12 farm is unsupported!");
        }
        public override bool ImportKeyword(
            TKeyword keywordElement, string searchCollectionUrl, 
            ProgramOptions.ConflictBehaviorType conflictBehavior, out int createdBestbets, out int createdSynonyms)
        {
            throw new NotSupportedException("Importing into an O12 farm is unsupported!");
        }
        public override bool ImportSearchSettings(
            TSearchSettings Element, string searchCollectionUrl,
            ProgramOptions.ConflictBehaviorType conflictBehavior)
        {
            throw new NotSupportedException("Importing into an O12 farm is unsupported!");
        }

        /// <summary>
        /// Lists the site collections known in the local farm
        /// </summary>
        /// <returns></returns>
        private System.Collections.IEnumerable ListSiteCollections()
        {
            SPFarm localFarm = SPFarm.Local;
            if (null == localFarm)
            {
                throw new ActionException("Cannot retrieve the a reference to the local farm. Make sure you are running this as a farm administrator.");
            }
            SPWebServiceCollection webServices = new SPWebServiceCollection(localFarm);

            foreach (SPWebService webService in webServices)
            {
                foreach (SPWebApplication webApp in webService.WebApplications)
                {
                    if (!webApp.IsAdministrationWebApplication)
                    {
                        foreach (SPSite siteCollection in webApp.Sites)
                        {
                            yield return siteCollection;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Lists the search services applications on the current farm
        /// </summary>
        /// <returns></returns>
        private static System.Collections.IEnumerable ListSearchServices()
        {
            SPFarm localFarm = SPFarm.Local;
            if (null == localFarm)
            {
                Logger.Log(TraceEventType.Error,"Cannot retrieve a reference to the local farm. Search applications will not be processed.");
                yield break;
            }

            foreach (SPServer spServer in localFarm.Servers)
            {
                SearchServiceInstanceCollection searchServices = new SearchServiceInstanceCollection(spServer);
                foreach (SearchServiceInstance searchService in searchServices)
                {
                    yield return searchService.Service as SearchService;
                }
            }
        }

        private class SearchServiceInstanceCollection : SPPersistedChildCollection<SearchServiceInstance>
        {
            public SearchServiceInstanceCollection(
                SPServer server)
                : base(server)
            {
            }
        }

        private void ExportKeywords(SPSite siteCollection, ExportAction exportAction)
        {
            try
            {
                SearchContext searchContext = SearchContext.GetContext(siteCollection);
                Keywords keywords = new Keywords(searchContext, new Uri(siteCollection.Url));
                if (keywords.AllKeywords.Count > 0) //only include in the output sitecollections that have keywords/bestbets
                {
                    ProgressIndicator keywordProgress =
                        new ProgressIndicator(
                            string.Format("Exporting keywords for Site [{0}]", siteCollection.Url),
                            keywords.AllKeywords.Count);

                    foreach (Keyword key in keywords.AllKeywords)
                    {
                        try
                        {
                            exportAction.WriteKeyword(siteCollection.Url, O12Conversions.ToKeyword(key));
                        }
                        catch (Exception exception)
                        {
                            Logger.Log(TraceEventType.Error, exception,
                                "Could not export Keyword [{0}] for Site [{1}].",
                                key.Term, siteCollection.Url);
                        }
                        finally
                        {
                            keywordProgress.IncreaseCounter();
                        }

                    }
                }
            }
            catch (Exception exception)
            {
                Logger.Log(TraceEventType.Error, exception, "Could not export Keywords for Site [{0}].", siteCollection.Url);
            }
        }

        private void ExportSiteScopesInformation(SPSite siteCollection, ExportAction exportAction)
        {
            try
            {
                SearchContext searchContext = SearchContext.GetContext(siteCollection);
                Scopes scopesManager = new Scopes(searchContext);
                Uri siteCollectionUri = new Uri(siteCollection.Url);

                ExportScopesForSite(exportAction, siteCollection, scopesManager);
                ExportGroupsForSite(exportAction, siteCollection, scopesManager);
            }
            catch (Exception exception)
            {
                Logger.Log(TraceEventType.Error, exception, "Could not export scope information for Site [{0}].", siteCollection.Url);
            }
        }

        private static void ExportGroupsForSite(ExportAction exportAction, SPSite siteCollection, Scopes scopesManager)
        {
            try
            {
                List<ScopeDisplayGroup> groupsForSite = Util.IEnumeratorToList<ScopeDisplayGroup>(
                    scopesManager.GetDisplayGroupsForSite(new Uri(siteCollection.Url)));
                if (groupsForSite.Count > 0) //only include in the output sitecollections that have scopes
                {
                    ProgressIndicator progress =
                        new ProgressIndicator(
                            string.Format("Exporting display groups for Site [{0}]", siteCollection.Url),
                            groupsForSite.Count);

                    foreach (ScopeDisplayGroup group in groupsForSite)
                    {
                        try
                        {
                            if (group.DisplayInAdminUI)
                            {
                                exportAction.WriteScopeDisplayGroup(siteCollection.Url, O12Conversions.ToScopeDisplayGroup(group));
                            }
                        }
                        catch (Exception exception)
                        {
                            Logger.Log(TraceEventType.Error, exception, "Could not export display group [{0}] for Site [{1}].", 
                                group.Name,siteCollection.Url);
                        }
                        finally
                        {
                            progress.IncreaseCounter();
                        }
                    }
                }
            }
            catch (Exception exception)
            {
                Logger.Log(TraceEventType.Error, exception, "Could not export display groups for Site [{0}].", siteCollection.Url);
            }
        }

        private static void ExportScopesForSite(ExportAction exportAction, SPSite siteCollection, Scopes scopesManager)
        {
            try
            {
                //creates lists with the scopes and groups of the site, so that we
                // can count them to do progress reporting
                List<Scope> scopesForSite = Util.IEnumeratorToList<Scope>(
                    scopesManager.GetScopesForSite(new Uri(siteCollection.Url)));
                if (scopesForSite.Count > 0) //only include in the output sitecollections that have scopes
                {
                    ProgressIndicator progress =
                        new ProgressIndicator(
                            string.Format("Exporting scopes for Site [{0}]", siteCollection.Url),
                            scopesForSite.Count);

                    foreach (Scope scope in scopesForSite)
                    {
                        try
                        {
                            if (scope.DisplayInAdminUI)
                            {
                                exportAction.WriteScope(siteCollection.Url, true, O12Conversions.ToScope(scope));
                            }
                        }
                        catch (Exception exception)
                        {
                            Logger.Log(TraceEventType.Error, exception,
                                "Could not export Scope [{0}] for Site [{1}].",
                                scope.Name, siteCollection.Url);
                        }
                        finally
                        {
                            progress.IncreaseCounter();
                        }
                    }
                }
            }
            catch (Exception exception)
            {
                Logger.Log(TraceEventType.Error, exception, "Could not scopes information for Site [{0}].", siteCollection.Url);
            }
        }

        private void ExportSharedScopesInformation(string searchAppName, SearchContext searchContext, ExportAction exportAction)
        {
            try
            {
                Scopes scopesManager = new Scopes(searchContext);
                List<Scope> scopesForSearchService = Util.IEnumeratorToList<Scope>(scopesManager.GetSharedScopes());

                if (scopesForSearchService.Count > 0) //only include in the output sitecollections that have scopes
                {
                    ProgressIndicator progress =
                        new ProgressIndicator(
                            string.Format("Exporting scopes for Search App [{0}]", searchAppName),
                            scopesForSearchService.Count);

                    foreach (Scope scope in scopesForSearchService)
                    {
                        try
                        {
                            if (scope.DisplayInAdminUI)
                            {
                                exportAction.WriteScope(searchAppName, false, O12Conversions.ToScope(scope));
                            }
                        }
                        catch (Exception exception)
                        {
                            Logger.Log(TraceEventType.Error, exception,
                                "Could not export Scope [{0}] for Search App [{1}].",
                                scope.Name, searchAppName);
                        }
                        finally
                        {
                            progress.IncreaseCounter();
                        }
                    }
                }
            }
            catch(Exception exception)
            {
                Logger.Log(TraceEventType.Error, exception, "Could not export Scopes for Search App [{0}].",searchAppName);
            }
        }
        public override bool ImportManagedProperties(TManagedProperty managedPropertyElement, string entityIdentifier, ProgramOptions.ConflictBehaviorType conflictBehavior, out int createdProperties, out int updatedProperties) {
            throw new NotImplementedException("Not supported by the O12 implementation!");
        }
    }

}