﻿using System;
using System.Collections.Generic;
using System.Reflection;
using System.Diagnostics;
using ArrayList = System.Collections.ArrayList;

using Microsoft.Office.Server.Search.Administration;


using Common;

namespace Layers
{
    public class O14Conversions
    {
        public static TKeyword ToKeyword(
            Keyword key, 
            Dictionary<Uri, BestBet> bestbets, 
            ArrayList bestbetOrder,
            Dictionary<string,Synonym> synonyms)
        {
            TKeyword newKey = new TKeyword();
            newKey.Contact = key.Contact;
            newKey.Definition = key.Definition;
            newKey.EndDate = key.EndDate;
            newKey.ReviewDate = key.ReviewDate;
            newKey.StartDate = key.StartDate;
            newKey.Term = key.Term;

            //In order to preserve betbet ordering we must consider the list of integer id in bestbetOrder
            //The problem is that BebtBet.id is internal, so to get it we must use reflection mojo.
            FieldInfo idInfo = typeof(BestBet).GetField("id", BindingFlags.Instance | BindingFlags.NonPublic);
            Debug.Assert(null != idInfo);
            newKey.Bestbet = new TBestbet[bestbets.Count];
            foreach (BestBet bb in bestbets.Values)
            {
                int bbId = (int)idInfo.GetValue(bb);
                int position = bestbetOrder.LastIndexOf(bbId);
                newKey.Bestbet[position] = ToBestbet(bb);
            }

            newKey.Synonym = new String[synonyms.Count];
            int i = 0;
            foreach (Synonym syn in synonyms.Values)
            {
                newKey.Synonym[i] = syn.Term;
                i++;
            }

            return newKey;
        }

        public static TBestbet ToBestbet(BestBet bb)
        {
            TBestbet newBB = new TBestbet();
            newBB.Description = bb.Description;
            newBB.Title = bb.Title;
            newBB.Url = bb.Url.ToString();
            return newBB;
        }

        public static TScope ToScope(ScopeInfo scope, List<RuleInfo> rulesForScope)
        {
            TScope newScope = new TScope();
            newScope.Name = scope.Name;
            newScope.AlternateResultsPage = scope.AlternateResultsPage;
            newScope.CompilationType = Util.MarshalEnum<ScopeCompilationType, TScopeCompilation>(scope.CompilationType);
            newScope.Description = scope.Description;
            newScope.ScopeKey = scope.Id.ToString();

            List<TAllContentScopeRule> allContentScopeRuleList;
            List<TUrlScopeRule> urlScopeRuleList;
            List<TPropertyQueryScopeRule> propertyQueryScopeRuleList;

            BucketRules(rulesForScope, out allContentScopeRuleList, out urlScopeRuleList, out propertyQueryScopeRuleList);

            newScope.AllContentScopeRule = allContentScopeRuleList.ToArray();
            newScope.UrlScopeRule = urlScopeRuleList.ToArray();
            newScope.PropertyQueryScopeRule = propertyQueryScopeRuleList.ToArray();
            newScope.Filter = scope.Filter;

            return newScope;
        }

        public static TScopeDisplayGroup ToScopeDisplayGroup(DisplayGroupInfo group, List<string> childScopes, string defaultScope)
        {
            TScopeDisplayGroup newGroup = new TScopeDisplayGroup();
            newGroup.Name = group.Name;
            newGroup.Description = group.Description;
            newGroup.DefaultScopeKey = (!String.IsNullOrEmpty(defaultScope)? defaultScope : null);

            newGroup.ScopeKey = childScopes.ToArray();
            return newGroup;
        }

        public static TUrlScopeRule ToUrlScopeRule(RuleInfo rule)
        {
            TUrlScopeRule newRule = new TUrlScopeRule();
            newRule.FilterBehavior = Util.MarshalEnum<ScopeRuleFilterBehavior, TFilterBehavior>(rule.FilterBehavior);
            newRule.UrlScopeRuleType = Util.MarshalEnum<UrlScopeRuleType, TUrlScopeRuleType>(rule.UrlRuleType);
            newRule.MatchingString = rule.UserValue;
            return newRule;
        }

        public static TPropertyQueryScopeRule ToPropertyQueryScopeRule(RuleInfo rule)
        {
            TPropertyQueryScopeRule newRule = new TPropertyQueryScopeRule();
            newRule.FilterBehavior = Util.MarshalEnum<ScopeRuleFilterBehavior, TFilterBehavior>(rule.FilterBehavior);
            newRule.Property = rule.ManagedProperty.Name;
            newRule.Value = rule.UserValue;
            return newRule;
        }

        private static void BucketRules(List<RuleInfo> rules,
           out List<TAllContentScopeRule> allContentScopeRuleList,
           out List<TUrlScopeRule> urlScopeRuleList,
           out List<TPropertyQueryScopeRule> propertyQueryScopeRuleList)
        {
            allContentScopeRuleList = new List<TAllContentScopeRule>();
            urlScopeRuleList = new List<TUrlScopeRule>();
            propertyQueryScopeRuleList = new List<TPropertyQueryScopeRule>();

            foreach (RuleInfo rule in rules)
            {
                switch (rule.RuleType)
                {
                    case ScopeRuleType.AllContent:
                        allContentScopeRuleList.Add(new TAllContentScopeRule());
                        break;
                    case ScopeRuleType.PropertyQuery:
                        propertyQueryScopeRuleList.Add(ToPropertyQueryScopeRule(rule));
                        break;
                    case ScopeRuleType.Url:
                        urlScopeRuleList.Add(ToUrlScopeRule(rule));
                        break;
                }
            }
        }
        public static TManagedProperty ToManagedProperty(ManagedProperty prop) {
            TManagedProperty mp = new TManagedProperty() {
                Name = prop.Name,
                Type= ToManagedPropertyType(prop.ManagedType),
                Description = prop.Description,
                EnabledForScoping = prop.EnabledForScoping,
                HasMultipleValues = prop.HasMultipleValues,
                FullTextQueriable = prop.FullTextQueriable
            };
            var crawledProperties = prop.GetMappedCrawledProperties(int.MaxValue);
            mp.CrawledProperty = new TCrawledProperty[crawledProperties.Count];
            int i = 0;
            foreach (var cp in crawledProperties) {
                mp.CrawledProperty[i] = new TCrawledProperty() { 
                    Name = cp.Name,
                    Propset = cp.Propset.ToString(),
                    VariantType = cp.VariantType,
                    Order = i++
                };
            }
            return mp;
        }

        private static TManagedPropertyType ToManagedPropertyType(ManagedDataType managedDataType) {
            switch (managedDataType) {
                case ManagedDataType.Text: return TManagedPropertyType.Text;
                case ManagedDataType.Binary: return TManagedPropertyType.Binary;
                case ManagedDataType.DateTime: return TManagedPropertyType.DateTime;
                case ManagedDataType.Decimal: return TManagedPropertyType.Decimal;
                case ManagedDataType.Integer: return TManagedPropertyType.Integer;
                case ManagedDataType.YesNo: return TManagedPropertyType.YesNo;
                case ManagedDataType.Unsupported:
                default:
                    return TManagedPropertyType.Unsupported;


            }
        }
        public static ManagedDataType ToManagedPropertyType(TManagedPropertyType type) {
            switch (type) {
                case TManagedPropertyType.Text: return ManagedDataType.Text;
                case TManagedPropertyType.Binary: return ManagedDataType.Binary;
                case TManagedPropertyType.DateTime: return ManagedDataType.DateTime;
                case TManagedPropertyType.Decimal: return ManagedDataType.Decimal;
                case TManagedPropertyType.Integer: return ManagedDataType.Integer;
                case TManagedPropertyType.YesNo: return ManagedDataType.YesNo;
                case TManagedPropertyType.Unsupported:
                default:
                    return ManagedDataType.Unsupported;


            }
        }

    }
}