﻿using System;
using System.Collections.Generic;
using ArrayList = System.Collections.ArrayList;
using System.Diagnostics;

using Microsoft.SharePoint;
using Microsoft.SharePoint.Administration;
using Microsoft.Office.Server.Search.Administration;

using Common;

namespace Layers {
    public class O14Layer : MossLayer {
        /// <summary>
        /// this contains the relationship between the scopekeys found on the 
        /// input file and the ids of the existent or created scopes in the 
        /// destiny farm. It is written when the scopes are processed and read
        /// when the displaygroups are processed. It is used to figure out 
        /// which scope in the destiny the displaygroup is refering to
        /// </summary>
        private Dictionary<string, int> scopeKeyToId = new Dictionary<string, int>();

        private string currentSiteCollectionUrl;
        private KeywordCollection currentSiteCollectionKeywordCollection;
        private string searchCenterUrl = string.Empty;
        private string dropdownMode = string.Empty;
        private string searchResultsUrl = string.Empty;
        private ScopesManager currentSiteCollectionScopesManager;
        private const string searchCenterUrlString = "SRCH_ENH_FTR_URL";
        private const string dropdownString = "SRCH_SITE_DROPDOWN_MODE";
        private const string searchResultsUrlString = "SRCH_TRAGET_RESULTS_PAGE";
        private void RetrieveSearchSettings(string searchCollectionUrl) {
            if (!string.Equals(currentSiteCollectionUrl, searchCollectionUrl, StringComparison.InvariantCultureIgnoreCase)) {
                try {
                    using (SPSite site = new SPSite(searchCollectionUrl)) {
                        SPSecurity.RunWithElevatedPrivileges(delegate() {
                            SPWeb rootWeb = site.RootWeb;

                            if (null != rootWeb) {

                                searchCenterUrl = rootWeb.AllProperties[searchCenterUrlString] as string;
                                dropdownMode = rootWeb.AllProperties[dropdownString] as string;
                                searchResultsUrl = rootWeb.AllProperties[searchResultsUrlString] as string;
                            }
                        });

                    }
                }
                catch (Exception exception) {
                    Logger.Log(TraceEventType.Error, exception,
                        "Cannot import information for Site [{0}] because I could not get a reference to it.",
                        searchCollectionUrl);
                }
                currentSiteCollectionUrl = searchCollectionUrl;
            }

        }
        private void PrepareToUseSiteCollection(string searchCollectionUrl) {
            if (!string.Equals(currentSiteCollectionUrl, searchCollectionUrl, StringComparison.InvariantCultureIgnoreCase)) {
                try {
                    using (SPSite site = new SPSite(searchCollectionUrl)) {
                        SearchServiceApplicationProxy searchApp = GetSearchAppProxyForSite(site);
                        currentSiteCollectionKeywordCollection = (new Keywords(searchApp, new Uri(searchCollectionUrl))).AllKeywords;
                        currentSiteCollectionScopesManager = new RemoteScopes(SPServiceContext.GetContext(site));

                    }
                }
                catch (Exception exception) {
                    Logger.Log(TraceEventType.Error, exception,
                        "Cannot import information for Site [{0}] because I could not get a reference to it.",
                        searchCollectionUrl);
                    currentSiteCollectionKeywordCollection = null;
                    currentSiteCollectionScopesManager = null;
                }
                currentSiteCollectionUrl = searchCollectionUrl;
            }
        }

        private SearchServiceApplication currentSearchApp;
        private string currentSearchAppName;
        private ScopesManager currentSearchAppScopesManager;
        private void PrepareToUseSearchApp(string searchAppName) {
            if (!string.Equals(currentSearchAppName, searchAppName, StringComparison.InvariantCultureIgnoreCase)) {
                currentSearchApp = GetSearchAppByName(searchAppName);
                if (currentSearchApp == null) {
                    Logger.Log(TraceEventType.Error,
                        "Cannot import information for Search App [{0}] an application with that name does not exist on the farm.",
                        searchAppName);
                }
                else {
                    currentSearchAppScopesManager = new Scopes(currentSearchApp);
                }
                currentSearchAppName = searchAppName;
            }
        }

        public override void ExportDefinitions(ExportAction exportAction) {

            if ((exportAction.Options.Entities & ProgramOptions.ActionEntities.Scope) != 0) {
                List<string> exported = new List<string>();    
                foreach (SearchServiceApplicationProxy searchApp in ListSearchApplications()) {
                    //ignores site collections that don't match the filter
                    if (!exportAction.ShouldProcessSearchApplication(searchApp.Name)) {
                        continue;
                    }
                    // ignore already exported ones, they can appear multiple times if re-used in proxy groups
                    if (exported.Contains(searchApp.Name)) {
                        continue;
                    }
                    ExportSharedScopesInformation(searchApp, exportAction);
                    exported.Add(searchApp.Name);
                }
            }

            if ((exportAction.Options.Entities & ProgramOptions.ActionEntities.ManagedProperties) != 0) {
                List<string> exported = new List<string>();    
                foreach (SearchServiceApplicationProxy searchApp in ListSearchApplications()) {
                    //ignores site collections that don't match the filter
                    if (!exportAction.ShouldProcessSearchApplication(searchApp.Name)) {
                        continue;
                    }
                    // ignore already exported ones, they can appear multiple times if re-used in proxy groups
                    if (exported.Contains(searchApp.Name)) {
                        continue;
                    }
                    ExportManagedProperties(searchApp, exportAction);
                    exported.Add(searchApp.Name);
                }
            }



            foreach (SPSite siteCollection in ListSiteCollections()) {
                if (siteCollection == null)
                    continue;

                //ignores site collections that don't match the filter
                if (!exportAction.ShouldProcessSiteCollection(siteCollection.Url)) {
                    continue;
                }

                if ((exportAction.Options.Entities & ProgramOptions.ActionEntities.Bestbet) != 0) {
                    ExportKeywordsInformation(siteCollection, exportAction);
                }

                if ((exportAction.Options.Entities & ProgramOptions.ActionEntities.Scope) != 0) {
                    ExportSiteScopesInformation(siteCollection, exportAction);
                }

                if ((exportAction.Options.Entities & ProgramOptions.ActionEntities.SearchSettings) != 0) {
                    ExportSearchSettingsInformation(siteCollection, exportAction);
                }
            }
        }

        public override bool ImportScopeDisplayGroup(TScopeDisplayGroup groupElement, string siteCollectionUrl, ProgramOptions.ConflictBehaviorType conflictBehavior) {
            bool createdGroup = false;

            PrepareToUseSiteCollection(siteCollectionUrl);
            if (currentSiteCollectionScopesManager == null) {
                return false;
            }

            Uri siteCollectionUri = new Uri(siteCollectionUrl);
            ScopeDisplayGroup displayGroup;
            bool shouldOverwrite = false;
            try {
                displayGroup = currentSiteCollectionScopesManager.GetDisplayGroup(siteCollectionUri, groupElement.Name);
                if (displayGroup != null) {   //we have a conflict since the group already exists
                    shouldOverwrite = ProgramOptions.HandleConflict(
                        conflictBehavior,
                        String.Format("Display Group [{0}]", groupElement.Name));
                }
            }
            catch (Exception) {
                displayGroup = null;
            }
            if (null != displayGroup && shouldOverwrite) {
                try {
                    try {
                        displayGroup.Delete();
                        displayGroup = null;
                    }
                    catch (InvalidOperationException exception) //display group is undeletable, i.e. built in
                    {
                        //in this case we clear the display group so that it can
                        //be filled by the scopes that come from the input file
                        while (displayGroup.Count > 0) {
                            displayGroup.RemoveAt(0);
                        }
                        displayGroup.Description = groupElement.Description;
                        displayGroup.Update();
                        createdGroup = true;

                        Logger.Log(TraceEventType.Information, exception,
                        "Cleared Display Group [{0}] for Site [{1}]. Now it has {2} scopes",
                        groupElement.Name,
                        siteCollectionUrl,
                        displayGroup.Count);
                    }
                }
                catch (Exception exception) {
                    Logger.Log(TraceEventType.Error, exception,
                        "Cannot delete Display Group [{0}] for Site [{1}].",
                        groupElement.Name,
                        siteCollectionUrl);
                    return false;
                }
            }
            if (null == displayGroup) {
                try {
                    displayGroup = currentSiteCollectionScopesManager.AllDisplayGroups.Create(
                        groupElement.Name, groupElement.Description, siteCollectionUri, true);
                    createdGroup = true;
                }
                catch (Exception exception) {
                    Logger.Log(TraceEventType.Error, exception,
                        "Cannot create Display Group [{0}] for Site [{1}].",
                        groupElement.Name,
                        siteCollectionUrl);
                    return false;
                }
            }
            else if (null != displayGroup && IsBuiltIn(displayGroup) && !shouldOverwrite) {
                //Special case: even in continue mode we add scopes to an existent display group if it is built in. 
                //This is different from overwriting because we just add scopes that are not already in the group 
                //to the end of it, instead of clearing the group first. We call this "merging".
                MergeInto(groupElement, displayGroup, siteCollectionUrl);
            }

            if (createdGroup) {
                FillDisplayGroup(groupElement, siteCollectionUrl, displayGroup);
            }
            else if (!createdGroup && !shouldOverwrite) {
                Logger.Log(TraceEventType.Error,
                     "Skipped Display Group [{0}] in Site [{1}] in the input file because it already existed. " +
                     "You can change the treatment of such conflicts with option conflictBehavior",
                     groupElement.Name, siteCollectionUrl);
            }
            return createdGroup;
        }

        public override bool ImportScope(TScope scopeElement, string entityIdentifier, bool isSiteScope, ProgramOptions.ConflictBehaviorType conflictBehavior, out int createdRules) {
            bool createdScope = false;
            createdRules = 0;
            ScopesManager scopesManager;
            string entity;
            if (isSiteScope) {
                PrepareToUseSiteCollection(entityIdentifier);
                scopesManager = currentSiteCollectionScopesManager;
                entity = String.Format("Site [{0}]", entityIdentifier);
            }
            else {
                PrepareToUseSearchApp(entityIdentifier);
                scopesManager = currentSearchAppScopesManager;
                entity = String.Format("Search App [{0}]", entityIdentifier);
            }
            if (scopesManager == null) {
                return false;
            }
            Scope scope = null;
            bool shouldAddRules = false;
            if (isSiteScope && currentSiteCollectionScopesManager != null) {
                scope = ImportSiteScope(scopeElement, scopesManager, conflictBehavior, out shouldAddRules);
            }
            else if (!isSiteScope && currentSearchAppScopesManager != null) {
                scope = ImportSharedScope(scopeElement, scopesManager, conflictBehavior, out shouldAddRules);
            }
            if (scope != null) {
                scopeKeyToId.Add(scopeElement.ScopeKey, scope.ID);
                createdScope = true;
            }

            if (scope != null && shouldAddRules) {
                // returns null if fastSearch switch is not specified
                string filter = GenerateScopeFilter(scopeElement, currentSearchApp, entityIdentifier);

                if (!String.IsNullOrEmpty(filter) && String.IsNullOrEmpty(scope.Filter)) {
                    scope.Filter = filter;
                    scope.Update();
                    Logger.Log(TraceEventType.Information, "Built FAST scope filter for scope [{0}] in entity [{1}]: [{2}]",
                        scopeElement.Name, entityIdentifier, filter);
                    createdRules++;
                }
                else {
                    createdRules = ImportRulesForScope(scopesManager, scope, scopeElement, entity);
                }
            }
            return createdScope;
        }

        public override bool ImportKeyword(TKeyword keywordElement, string siteCollectionUrl, ProgramOptions.ConflictBehaviorType conflictBehavior, out int createdBestbets, out int createdSynonyms) {
            bool createdKeyword = false;
            createdBestbets = 0;
            createdSynonyms = 0;
            PrepareToUseSiteCollection(siteCollectionUrl);
            if (currentSiteCollectionKeywordCollection == null) {
                return false;
            }

            Keyword keyword;
            bool shouldOverwrite = false;
            try {
                keyword = currentSiteCollectionKeywordCollection[keywordElement.Term];
                if (null != keyword) {
                    //we have a conflict since the keyword already exists
                    shouldOverwrite = ProgramOptions.HandleConflict(
                        conflictBehavior,
                        String.Format("Keyword [{0}]", keywordElement.Term));
                }
            }
            catch (Exception) {
                keyword = null;
            }
            if (null != keyword && shouldOverwrite) {
                try {
                    keyword.Delete();
                    Logger.Log(TraceEventType.Information,
                       "Deleted Keyword [{0}] for [{1}].",
                       keywordElement.Term,
                       siteCollectionUrl);
                    keyword = null;
                }
                catch (Exception exception) {
                    Logger.Log(TraceEventType.Error, exception,
                        "Cannot delete Keyword [{0}] for [{1}].",
                        keywordElement.Term,
                        siteCollectionUrl);
                    return false;
                }
            }
            if (null == keyword) {
                try {
                    keyword = currentSiteCollectionKeywordCollection.Create(keywordElement.Term, keywordElement.StartDate);
                    createdKeyword = true;
                    try {
                        keyword.Contact = keywordElement.Contact;
                        keyword.EndDate = keywordElement.EndDate;
                        keyword.ReviewDate = keywordElement.ReviewDate;
                        keyword.Definition = keywordElement.Definition;
                        keyword.Update();
                    }
                    catch (Exception exception) {
                        Logger.Log(TraceEventType.Error, exception,
                            "Cannot complete Keyword [{0}] in Site [{1}].",
                            keywordElement.Term, siteCollectionUrl);
                    }
                }
                catch (Exception exception) {
                    Logger.Log(TraceEventType.Error, exception,
                        "Cannot create Keyword [{0}] in Site [{1}].",
                        keywordElement.Term, siteCollectionUrl);
                    return false;
                }

                if (keywordElement.Synonym != null) {
                    foreach (string synonymElement in keywordElement.Synonym) {
                        try {
                            keyword.Synonyms.Create(synonymElement);
                            createdSynonyms++;
                        }
                        catch (ArgumentException exception) {
                            Logger.Log(TraceEventType.Error, exception,
                               "Cannot create Synonym [{0}] for Keyword [{1}] in Site [{2}].",
                               synonymElement, keywordElement.Term, siteCollectionUrl);
                            continue;
                        }
                    }
                }
                createdBestbets += ImportBestBets(keywordElement, keyword, conflictBehavior);
            }
            else if (conflictBehavior != ProgramOptions.ConflictBehaviorType.Prompt) {
                Logger.Log(TraceEventType.Error,
                     "Skipped Keyword [{0}] in Site [{1}] in the input file because it already existed. " +
                     "You can change the treatment of such conflicts with option conflictBehavior",
                     keywordElement.Term, siteCollectionUrl);
            }
            return createdKeyword;
        }


        public override bool ImportSearchSettings(TSearchSettings element, string siteCollectionUrl,
            ProgramOptions.ConflictBehaviorType conflictBehavior) {
            RetrieveSearchSettings(siteCollectionUrl);
            bool shouldOverwrite = false;

            if (string.IsNullOrEmpty(element.SearchCenterUrl) && string.IsNullOrEmpty(element.DropdownMode.ToString()) && string.IsNullOrEmpty(element.SearchResultsUrl)) {
                return false;
            }
            //before , it is string.nullorempty
            if (!string.IsNullOrEmpty(element.SearchCenterUrl)) {
                try {

                    if (!string.IsNullOrEmpty(searchCenterUrl)) {
                        //we have a conflict since the SearchCenterUrl already exists
                        shouldOverwrite = ProgramOptions.HandleConflict(
                            conflictBehavior,
                            String.Format("SearchCenterUrl [{0}]", searchCenterUrl));
                    }
                }
                catch (Exception) {
                    Logger.Log(TraceEventType.Error, "skipped conflict settings");

                }
                Logger.Log(TraceEventType.Information, "should overwrite searchcenter url {0}", shouldOverwrite);

                if ((!string.IsNullOrEmpty(searchCenterUrl) && shouldOverwrite) || string.IsNullOrEmpty(searchCenterUrl)) {
                    try {

                        Logger.Log(TraceEventType.Information,
                           "overwrite SearchCenterUrl [{0}] for [{1}].",
                           element.SearchCenterUrl,
                           siteCollectionUrl);
                        using (SPSite site = new SPSite(siteCollectionUrl)) {
                            SPSecurity.RunWithElevatedPrivileges(delegate() {
                                SPWeb rootWeb = site.RootWeb;

                                if (null != rootWeb) {
                                    rootWeb.AllProperties[searchCenterUrlString] = element.SearchCenterUrl;
                                    rootWeb.Update();
                                }
                            });

                        }
                    }
                    catch (Exception exception) {
                        Logger.Log(TraceEventType.Error, exception,
                            "Cannot overwrite SearchCenterUrl [{0}] for [{1}].",
                            element.SearchCenterUrl,
                            siteCollectionUrl);
                    }
                }
            }

            if (!string.IsNullOrEmpty(element.DropdownMode.ToString())) {
                shouldOverwrite = false;
                try {

                    if (!string.IsNullOrEmpty(dropdownMode)) {
                        //we have a conflict since the keyword already exists
                        shouldOverwrite = ProgramOptions.HandleConflict(
                            conflictBehavior,
                            String.Format("DropdownMode [{0}]", dropdownMode));
                    }
                }
                catch (Exception) {

                }
                if ((!string.IsNullOrEmpty(dropdownMode) && shouldOverwrite) || string.IsNullOrEmpty(dropdownMode)) {
                    try {

                        Logger.Log(TraceEventType.Information,
                           "overwrite DropdownMode [{0}] for [{1}].",
                           element.DropdownMode.ToString(),
                           siteCollectionUrl);
                        using (SPSite site = new SPSite(siteCollectionUrl)) {
                            SPSecurity.RunWithElevatedPrivileges(delegate() {
                                SPWeb rootWeb = site.RootWeb;

                                if (null != rootWeb) {
                                    rootWeb.AllProperties[dropdownString] = element.DropdownMode.ToString();
                                    rootWeb.Update();
                                }
                            });

                        }
                    }
                    catch (Exception exception) {
                        Logger.Log(TraceEventType.Error, exception,
                            "Cannot overwrite DropdownMode [{0}] for [{1}].",
                            element.DropdownMode.ToString(),
                            siteCollectionUrl);
                        return false;
                    }
                }
            }

            if (!string.IsNullOrEmpty(element.SearchResultsUrl)) {
                shouldOverwrite = false;
                try {

                    if (!string.IsNullOrEmpty(searchResultsUrl)) {
                        //we have a conflict since the keyword already exists
                        shouldOverwrite = ProgramOptions.HandleConflict(
                            conflictBehavior,
                            String.Format("SearchResultsUrl [{0}]", searchResultsUrl));
                    }
                }
                catch (Exception) {

                }
                if ((!string.IsNullOrEmpty(searchResultsUrl) && shouldOverwrite) || string.IsNullOrEmpty(searchResultsUrl)) {
                    try {

                        Logger.Log(TraceEventType.Information,
                           "overwrite SearchResultsUrl [{0}] for [{1}].",
                           element.SearchResultsUrl,
                           siteCollectionUrl);
                        using (SPSite site = new SPSite(siteCollectionUrl)) {
                            SPSecurity.RunWithElevatedPrivileges(delegate() {
                                SPWeb rootWeb = site.RootWeb;

                                if (null != rootWeb) {
                                    rootWeb.AllProperties[searchResultsUrlString] = element.SearchResultsUrl;
                                    rootWeb.Update();
                                }
                            });

                        }
                    }
                    catch (Exception exception) {
                        Logger.Log(TraceEventType.Error, exception,
                            "Cannot overwrite DropdownMode [{0}] for [{1}].",
                            element.SearchResultsUrl,
                            siteCollectionUrl);
                        return false;
                    }
                }
            }
            return true;

        }

        /// <summary>
        /// Merges groupElement into displayGroup. 
        /// I.e Appends to displayGroup any scopes that are in groupElement 
        /// but not on displayGroup.
        /// </summary>
        /// <param name="groupElement"></param>
        /// <param name="displayGroup"></param>
        private void MergeInto(TScopeDisplayGroup groupElement, ScopeDisplayGroup displayGroup, string siteCollectionUrl) {
            if (groupElement.ScopeKey != null) {
                int addedScopes = 0;
                foreach (string scopeKay in groupElement.ScopeKey) {
                    Scope scope = GetScope(currentSiteCollectionScopesManager, scopeKay);
                    if (scope != null) {
                        if (!displayGroup.Contains(scope)) {
                            displayGroup.Add(scope);
                            addedScopes++;
                        }
                    }
                }
                if (addedScopes > 0) {
                    displayGroup.Update();
                    Logger.Log(TraceEventType.Information,
                        "Added {0} scopes to Display Group [{1}] in site collection [{2}] without overwriting it.",
                        addedScopes, groupElement.Name, siteCollectionUrl);
                }
            }
        }

        /// <summary>
        /// Only checks the name, so it only works for english site collections
        /// </summary>
        /// <param name="displayGroup"></param>
        /// <returns></returns>
        private bool IsBuiltIn(ScopeDisplayGroup displayGroup) {
            return
                String.Equals(displayGroup.Name, "Search Dropdown", StringComparison.InvariantCultureIgnoreCase) ||
                String.Equals(displayGroup.Name, "Advanced Search", StringComparison.InvariantCultureIgnoreCase);
        }

        /// <summary>
        /// Fills the display group with scopes and the default scope
        /// </summary>
        /// <param name="groupElement"></param>
        /// <param name="siteCollectionUrl"></param>
        /// <param name="displayGroup"></param>
        private void FillDisplayGroup(TScopeDisplayGroup groupElement, string siteCollectionUrl, ScopeDisplayGroup displayGroup) {
            if (null != groupElement.ScopeKey) {
                foreach (string scopeKey in groupElement.ScopeKey) {
                    try {
                        Scope scopeInGroup = GetScope(currentSiteCollectionScopesManager, scopeKey);
                        if (scopeInGroup != null) {
                            displayGroup.Add(scopeInGroup);
                        }
                        else {
                            throw new ActionException("Referenced Scope does not exist.");
                        }
                    }
                    catch (Exception exception) {
                        Logger.Log(TraceEventType.Error, exception,
                           "Cannot add Scope [key={0}] to Display Group [{1}] for Site [{2}].",
                           scopeKey, groupElement.Name, siteCollectionUrl);
                    }
                }
            }

            try {
                if (!string.IsNullOrEmpty(groupElement.DefaultScopeKey)) {
                    Scope defaultScope = GetScope(currentSiteCollectionScopesManager, groupElement.DefaultScopeKey); ;
                    if (defaultScope != null) {
                        displayGroup.Default = defaultScope;
                    }
                }
            }
            catch (Exception exception) {
                Logger.Log(TraceEventType.Error, exception,
                   "Cannot set Scope [key={0}] as Default in Display Group [{1}] in Site [{2}].",
                   groupElement.DefaultScopeKey, groupElement.Name, siteCollectionUrl);
            }
            displayGroup.Update();
        }

        private Scope GetScope(ScopesManager scopesManager, string scopeKey) {
            Scope result = null;
            //checks is the refered scope is a built in one
            if (Util.BuiltInScopeKeys.IsBuiltIn(scopeKey)) {
                return scopesManager.AllScopes[Util.BuiltInScopeKeys.KeyToId(scopeKey)];
            }

            int scopeId = 0;
            if (scopeKeyToId.TryGetValue(scopeKey, out scopeId)) {
                result = scopesManager.AllScopes[scopeId];
            }
            return result;
        }

        protected virtual string GenerateScopeFilter(TScope scopeElement, SearchServiceApplication ssa, string siteCollectionUrl) {
            // Filters are created in FS14Layer
            return null;
        }

        private Scope ImportSharedScope(TScope scopeElement, ScopesManager scopesManager, ProgramOptions.ConflictBehaviorType conflictBehavior, out bool shouldAddRules) {
            //when a scope is created by this function this will be set to true
            shouldAddRules = false;

            Scope scope = null;
            bool shouldOverwrite = false;
            try {
                if (!Util.BuiltInScopeKeys.IsBuiltIn(scopeElement.ScopeKey)) {
                    scope = scopesManager.GetSharedScope(scopeElement.Name);
                }
                else {
                    scope = scopesManager.AllScopes[Util.BuiltInScopeKeys.KeyToId(scopeElement.ScopeKey)];
                }

                if (scope != null) {   //we have a conflict since the scope already exists
                    shouldOverwrite = ProgramOptions.HandleConflict(
                        conflictBehavior,
                        String.Format("Scope [{0}]", scopeElement.Name));
                }
            }
            catch (Exception) {
                scope = null;
            }
            if (null != scope && shouldOverwrite) {
                if (!Util.BuiltInScopeKeys.IsBuiltIn(scopeElement.ScopeKey)) {
                    try {
                        scope.Delete();
                        Logger.Log(TraceEventType.Information,
                            "Deleted Scope [{0}] for Search App [{1}].",
                            scopeElement.Name,
                            currentSearchAppName);
                        scope = null;
                    }
                    catch (Exception exception) {
                        Logger.Log(TraceEventType.Error, exception,
                            "Cannot delete Scope [{0}] for Search App [{1}].",
                            scopeElement.Name,
                            currentSearchAppName);
                        scope = null;
                    }
                }
                else //clear built in scope so that it can be rebuilt from the input file
                {
                    try {
                        //we cant delete the rules in place in the ScupeRulesCollection, 
                        //so we build a list of them and them iterate on that list to delete them
                        List<ScopeRule> rules = new List<ScopeRule>();
                        foreach (ScopeRule rule in scope.Rules) {
                            rules.Add(rule);
                        }
                        foreach (ScopeRule rule in rules) {
                            rule.Delete();
                        }
                        scope.AlternateResultsPage = !String.IsNullOrEmpty(scopeElement.AlternateResultsPage) ? scopeElement.AlternateResultsPage : null;
                        scope.CompilationType = Util.MarshalEnum<TScopeCompilation, ScopeCompilationType>(scopeElement.CompilationType);
                        scope.Name = scopeElement.Name;
                        scope.Description = scopeElement.Description;
                        scope.Update();
                        shouldAddRules = true;
                    }
                    catch (Exception exception) {
                        Logger.Log(TraceEventType.Error, exception,
                          "Cannot clear Scope [{0}] for Search App [{1}].",
                          scopeElement.Name,
                          currentSearchAppName);
                    }
                }
            }
            else if (null != scope && !shouldOverwrite) {
                Logger.Log(TraceEventType.Error,
                     "Skipped Scope [{0}] in Search App [{1}] in the input file because it already existed. " +
                     "You can change the treatment of such conflicts with option conflictBehavior",
                     scopeElement.Name, currentSearchAppName);
                scopeKeyToId.Add(scopeElement.ScopeKey, scope.ID);
                return null;
            }

            if (null == scope) {
                try {
                    scope = scopesManager.AllScopes.Create(
                        scopeElement.Name, scopeElement.Description,
                        null, true,
                        !String.IsNullOrEmpty(scopeElement.AlternateResultsPage) ? scopeElement.AlternateResultsPage : null,
                        Util.MarshalEnum<TScopeCompilation, ScopeCompilationType>(scopeElement.CompilationType),
                        String.IsNullOrEmpty(scopeElement.Filter) ? null : scopeElement.Filter);
                    shouldAddRules = true;
                    return scope;
                }
                catch (Exception exception) {
                    Logger.Log(TraceEventType.Error, exception,
                        "Cannot create Scope [{0}] for Search App [{1}].",
                        scopeElement.Name,
                        currentSearchAppName);
                    scope = null;
                }
            }
            return scope;
        }

        private Scope ImportSiteScope(TScope scopeElement, ScopesManager scopesManager, ProgramOptions.ConflictBehaviorType conflictBehavior, out bool shouldAddRules) {
            shouldAddRules = false;
            Uri siteCollectionUri = new Uri(currentSiteCollectionUrl);
            Scope scope = null;
            bool shouldOverwrite = false;
            try {
                scope = scopesManager.GetScope(siteCollectionUri, scopeElement.Name);
                if (scope != null) {   //we have a conflict since the scope already exists
                    shouldOverwrite = ProgramOptions.HandleConflict(
                        conflictBehavior,
                        String.Format("Scope [{0}]", scopeElement.Name));
                }
            }
            catch (Exception) {
                scope = null;
            }
            if (null != scope && shouldOverwrite) {
                try {
                    scope.Delete();
                    Logger.Log(TraceEventType.Information,
                        "Delete Scope [{0}] for Site [{1}].",
                        scopeElement.Name,
                        currentSiteCollectionUrl);
                    scope = null;
                }
                catch (Exception exception) {
                    Logger.Log(TraceEventType.Error, exception,
                        "Cannot delete Scope [{0}] for Site [{1}].",
                        scopeElement.Name,
                        currentSiteCollectionUrl);
                    scope = null;
                }
            }
            if (null == scope) {
                try {
                    scope = scopesManager.AllScopes.Create(
                        scopeElement.Name, scopeElement.Description,
                        siteCollectionUri, true,
                        !String.IsNullOrEmpty(scopeElement.AlternateResultsPage) ? scopeElement.AlternateResultsPage : null,
                        Util.MarshalEnum<TScopeCompilation, ScopeCompilationType>(scopeElement.CompilationType),
                        String.IsNullOrEmpty(scopeElement.Filter) ? null : scopeElement.Filter);
                    shouldAddRules = true;
                    return scope;
                }
                catch (Exception exception) {
                    Logger.Log(TraceEventType.Error, exception,
                        "Cannot create Scope [{0}] for Site [{1}].",
                        scopeElement.Name,
                        currentSiteCollectionUrl);
                    scope = null;
                }
            }
            else if (null != scope && !shouldOverwrite) {
                Logger.Log(TraceEventType.Error,
                     "Skipped Scope [{0}] in Site [{1}] in the input file because it already existed. " +
                     "You can change the treatment of such conflicts with option conflictBehavior",
                     scopeElement.Name, currentSiteCollectionUrl);
                scopeKeyToId.Add(scopeElement.ScopeKey, scope.ID);
                scope = null;
            }
            return scope;
        }

        private static int ImportRulesForScope(ScopesManager scopesManager, Scope scope, TScope scopeElement, string entity) {
            return
                ImportAllContentRules(scope, scopeElement, entity) +
                ImportUrlScopeRules(scope, scopeElement, entity) +
                ImportPropertyQueryRules(scopesManager, scope, scopeElement, entity);
        }

        private static int ImportPropertyQueryRules(ScopesManager scopesManager, Scope scope, TScope scopeElement, string entity) {
            int createdRules = 0;
            if (null != scopeElement.PropertyQueryScopeRule &&
                scopeElement.PropertyQueryScopeRule.Length > 0) {
                IDictionary<string, ManagedPropertyInfo> propertiesByName = null;
                try {
                    System.Reflection.PropertyInfo propertiesByNameType = scopesManager.GetType().GetProperty(
                        "ManagedPropertiesByName", System.Reflection.BindingFlags.Instance | System.Reflection.BindingFlags.NonPublic);
                    propertiesByName = propertiesByNameType.GetValue(scopesManager, null) as IDictionary<string, ManagedPropertyInfo>;
                    Debug.Assert(propertiesByName != null);
                }
                catch (Exception ex) {
                    Logger.Log(TraceEventType.Error, ex, "Cannot import PropertyRule for Scope [{0}] in [{1}].", scopeElement.Name, entity);
                }
                foreach (TPropertyQueryScopeRule rule in scopeElement.PropertyQueryScopeRule) {
                    try {
                        ManagedPropertyInfo property;
                        if (propertiesByName.TryGetValue(rule.Property, out property)) {
                            //The O12 defaut content source is named "Local Office SharePoint Server sites",
                            //whereas the O14 one is "Local SharePoint sites", here we do the conversion between the two, 
                            //so that we don't invalidate scopes that use it.
                            string ruleValue = rule.Value;
                            if (String.Equals(rule.Property, "ContentSource", StringComparison.InvariantCultureIgnoreCase) &&
                                String.Equals(ruleValue, "Local Office SharePoint Server sites", StringComparison.InvariantCultureIgnoreCase)) {
                                ruleValue = "Local SharePoint sites";
                            }

                            scope.Rules.CreatePropertyQueryRule(
                                Util.MarshalEnum<TFilterBehavior, ScopeRuleFilterBehavior>(rule.FilterBehavior),
                                property,
                                ruleValue);
                            createdRules++;
                        }
                        else {
                            Logger.Log(TraceEventType.Error,
                               "Cannot add Property Rule [{0}={1}] to Scope [{2}] in {3}. Managed Property [{0}] does not exist.",
                               rule.Property, rule.Value, scopeElement.Name, entity);
                        }
                    }
                    catch (Exception exception) {
                        Logger.Log(TraceEventType.Error, exception,
                             "Cannot add Property Rule [{0}={1}] to Scope [{2}] in {3}.",
                             rule.Property, rule.Value, scopeElement.Name, entity);
                    }
                }
            }
            return createdRules;
        }

        private static int ImportUrlScopeRules(Scope scope, TScope scopeElement, string entity) {
            int createdRules = 0;
            if (scopeElement.UrlScopeRule != null) {
                foreach (TUrlScopeRule rule in scopeElement.UrlScopeRule) {
                    try {
                        scope.Rules.CreateUrlRule(
                            Util.MarshalEnum<TFilterBehavior, ScopeRuleFilterBehavior>(rule.FilterBehavior),
                            Util.MarshalEnum<TUrlScopeRuleType, UrlScopeRuleType>(rule.UrlScopeRuleType),
                            rule.MatchingString);
                        createdRules++;
                    }
                    catch (Exception exception) {
                        Logger.Log(TraceEventType.Error, exception,
                          "Cannot create Url Rule [{0}, {1}] to Scope [{2}] in {3}.",
                          rule.UrlScopeRuleType.ToString(),
                          rule.MatchingString,
                          scopeElement.Name,
                          entity);
                    }
                }
            }
            return createdRules;
        }

        private static int ImportAllContentRules(Scope scope, TScope scopeElement, string entity) {
            int createdRules = 0;
            if (scopeElement.AllContentScopeRule != null) {
                foreach (TAllContentScopeRule rule in scopeElement.AllContentScopeRule) {
                    try {
                        scope.Rules.CreateAllContentRule();
                        createdRules++;
                    }
                    catch (Exception exception) {
                        Logger.Log(TraceEventType.Error, exception,
                          "Cannot add All Content Rule to Scope [{0}] in {1}.",
                          scopeElement.Name, entity);
                    }
                }
            }
            return createdRules;
        }

        private int ImportBestBets(TKeyword keywordElement, Keyword keyword, ProgramOptions.ConflictBehaviorType conflictBehavior) {
            int createdBestbets = 0;
            if (keywordElement.Bestbet != null) {
                foreach (TBestbet bestbetElement in keywordElement.Bestbet) {
                    BestBet bb;
                    bool shouldOverwrite = false;
                    try {
                        bb = keyword.BestBets[new Uri(bestbetElement.Url)];
                        //we have a conflict since the keyword already exists
                        shouldOverwrite = ProgramOptions.HandleConflict(
                            conflictBehavior,
                            String.Format("BestBet [{0}] for Keyword [{1}] in Site [{2}]", bestbetElement.Url, keyword.Term, currentSiteCollectionUrl));
                    }
                    catch (Exception) {
                        bb = null;
                    }
                    if (null != bb && shouldOverwrite) {
                        try {
                            keyword.BestBets.Remove(bb);
                            bb = null;
                        }
                        catch (Exception exception) {
                            Logger.Log(TraceEventType.Error, exception,
                                "Cannot delete BestBet [{0}] for Keyword [{1}] in Site [{2}].",
                                bestbetElement.Url,
                                keywordElement.Term,
                                currentSiteCollectionUrl);
                            continue;
                        }
                    }
                    else if (null == bb && !shouldOverwrite) {
                        try {
                            bb = keyword.BestBets.Create(
                                    bestbetElement.Title,
                                    bestbetElement.Description,
                                    new Uri(bestbetElement.Url));
                            createdBestbets++;
                        }
                        catch (Exception exception) {
                            Logger.Log(TraceEventType.Error, exception,
                               "Cannot create BestBet [{0}] for Keyword [{1}] in  Site [{2}].",
                               bestbetElement.Url, keywordElement.Term, currentSiteCollectionUrl);
                            continue;
                        }
                    }
                }
            }
            return createdBestbets;
        }

        /// <summary>
        /// Lists the site collections known in the local farm
        /// </summary>
        /// <returns></returns>
        private System.Collections.IEnumerable ListSiteCollections() {
            SPFarm localFarm = SPFarm.Local;
            if (null == localFarm) {
                throw new ActionException("Cannot retrieve the a reference to the local farm. Make sure you are running this as a farm administrator.");
            }
            SPWebServiceCollection webServices = new SPWebServiceCollection(localFarm);

            foreach (SPWebService webService in webServices) {
                foreach (SPWebApplication webApp in webService.WebApplications) {
                    if (!webApp.IsAdministrationWebApplication) {
                        foreach (SPSite siteCollection in webApp.Sites) {
                            yield return siteCollection;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Lists the search services applications on the current farm
        /// </summary>
        /// <returns></returns>
        private static System.Collections.IEnumerable ListSearchApplications() {
            foreach (SPServiceApplicationProxyGroup proxyGroup in SPFarm.Local.ServiceApplicationProxyGroups) {
                foreach (SPServiceApplicationProxy proxy in proxyGroup.Proxies) {
                    SearchServiceApplicationProxy searchApp = proxy as SearchServiceApplicationProxy;
                    if (searchApp == null) {
                        continue;
                    }
                    yield return searchApp;
                }
            }
        }

        protected static SearchServiceApplicationProxy GetSearchAppProxyForSite(SPSite siteCollection) {
            return (SearchServiceApplicationProxy)
                SearchServiceApplicationProxy.GetProxy(SPServiceContext.GetContext(siteCollection));
        }

        protected static SearchServiceApplicationProxy GetSearchAppProxyByName(string searchAppName) {
            foreach (SearchServiceApplicationProxy app in ListSearchApplications()) {
                if (String.Equals(app.Name, searchAppName, StringComparison.InvariantCultureIgnoreCase)) {
                    return app;
                }
            }
            return null;
        }

        protected static SearchServiceApplication GetSearchAppByName(string searchAppName) {
            return SearchService.Service.SearchApplications.GetValue<SearchServiceApplication>(searchAppName) as SearchServiceApplication;
        }

        private void ExportKeywordsInformation(SPSite siteCollection, ExportAction exportAction) {
            try {
                SearchServiceApplicationProxy searchApp = GetSearchAppProxyForSite(siteCollection);
                var keywordsForSite = searchApp.GetSpecialTermsInfo(
                    KeywordView.AllKeywords, KeywordFilter.None, null, siteCollection.ID.ToString());

                if (keywordsForSite.Count > 0) //only include in the output sitecollections that have keywords/bestbets
                {
                    ProgressIndicator keywordProgress =
                        new ProgressIndicator(
                            string.Format("Exporting keywords for Site [{0}]", siteCollection.Url),
                            keywordsForSite.Count);

                    foreach (Keyword key in keywordsForSite.Values) {
                        try {
                            ArrayList bestbetsOrder = new ArrayList(); ;
                            Dictionary<Uri, BestBet> bestbets = searchApp.GetBestBetsInfoForSpecialTerm(key, ref bestbetsOrder);
                            Dictionary<string, Synonym> synonyms = searchApp.GetSynonymsInfoSpecialTerm(key);
                            exportAction.WriteKeyword(siteCollection.Url, O14Conversions.ToKeyword(key, bestbets, bestbetsOrder, synonyms));
                        }
                        catch (Exception ex) {
                            Logger.Log(TraceEventType.Error, ex, "Could not export keyword [{0}] for site [{1}]", key.Term, siteCollection.Url);
                        }
                        finally {
                            keywordProgress.IncreaseCounter();
                        }
                    }
                }
            }
            catch (Exception ex) {
                Logger.Log(TraceEventType.Error, ex, "Could not export keywords for site [{0}]", siteCollection.Url);
            }
        }

        private void ExportSiteScopesInformation(SPSite siteCollection, ExportAction exportAction) {
            if (siteCollection == null) {
                Logger.Log(TraceEventType.Warning, "O14Layer::ExportScopes: siteCollection is null");
                return;
            }

            SearchServiceApplicationProxy searchApp = GetSearchAppProxyForSite(siteCollection);

            if (searchApp == null) {
                Logger.Log(TraceEventType.Warning, "Search Application Proxy is null for site {0}", siteCollection.ID);
                return;
            }

            ExportScopesForSite(exportAction, siteCollection, searchApp);
            ExportGroupsForSite(exportAction, siteCollection, searchApp);
        }

        private void ExportSearchSettingsInformation(SPSite siteCollection, ExportAction exportAction) {
            if (siteCollection == null) {
                Logger.Log(TraceEventType.Warning, "O14Layer::ExportSearchSettings: siteCollection is null");
                return;
            }
            ExportSearchSettingsForSite(exportAction, siteCollection);
        }

        private void ExportSearchSettingsForSite(ExportAction exportAction, SPSite siteCollection) {
            try {
                Uri siteCollectionUri = new Uri(siteCollection.Url);

                RetrieveSearchSettings(siteCollection.Url);

                ProgressIndicator progress =
                    new ProgressIndicator(
                        string.Format("Exporting search settings for Site [{0}]", siteCollection.Url),
                        1);

                exportAction.WriteSearchSettings(siteCollection.Url, searchCenterUrl, dropdownMode, searchResultsUrl);

                progress.IncreaseCounter();

            }
            catch (Exception ex) {
                Logger.Log(TraceEventType.Error, ex, "Could not export the search settings for site [{0}]", siteCollection.Url);
            }
        }

        private static void ExportScopesForSite(ExportAction exportAction, SPSite siteCollection, SearchServiceApplicationProxy searchApp) {
            try {
                Uri siteCollectionUri = new Uri(siteCollection.Url);

                //creates lists with the scopes, so that we
                // can count them to do progress reporting
                List<ScopeInfo> scopesForSite = searchApp.GetScopesForConsumer(siteCollection.ID.ToString());
                if (scopesForSite.Count > 0) //only include in the output sitecollections that have scopes
                {
                    ProgressIndicator progress =
                        new ProgressIndicator(
                            string.Format("Exporting scopes for Site [{0}]", siteCollection.Url),
                            scopesForSite.Count);

                    foreach (ScopeInfo scope in scopesForSite) {
                        try {
                            if (scope.DisplayInAdminUI) {
                                int statusCode;
                                List<RuleInfo> rulesForScope = searchApp.GetRulesInfo(scope.Id, out statusCode);
                                exportAction.WriteScope(siteCollection.Url, true, O14Conversions.ToScope(scope, rulesForScope));
                            }
                        }
                        catch (Exception ex) {
                            Logger.Log(TraceEventType.Error, ex, "Could not export scope [{0}] for site [{1}]",
                                scope.Name, siteCollection.Url);
                        }
                        finally {
                            progress.IncreaseCounter();
                        }
                    }
                }
            }
            catch (Exception ex) {
                Logger.Log(TraceEventType.Error, ex, "Could not export the scopes for site [{0}]", siteCollection.Url);
            }
        }

        private static void ExportGroupsForSite(ExportAction exportAction, SPSite siteCollection, SearchServiceApplicationProxy searchApp) {
            try {
                List<DisplayGroupInfo> groupsForSite = searchApp.GetDisplayGroupsForConsumer(siteCollection.ID.ToString());
                if (groupsForSite.Count > 0) //only include in the output sitecollections that have scopes
                {
                    ProgressIndicator progress =
                        new ProgressIndicator(
                            String.Format("Exporting display groups for Site [{0}]", siteCollection.Url),
                            groupsForSite.Count);

                    foreach (DisplayGroupInfo group in groupsForSite) {
                        try {
                            if (group.DisplayInAdminUI) {
                                List<int> scopesIdsForGroup = searchApp.GetDisplayGroupListInfo(group.Id);
                                List<string> scopesIndexesForGroup = new List<string>();
                                foreach (int id in scopesIdsForGroup) {
                                    scopesIndexesForGroup.Add(id.ToString());
                                }
                                string defaultScope = String.Empty;
                                if (group.DefaultScopeID != 0) {
                                    defaultScope = group.DefaultScopeID.ToString();
                                }

                                exportAction.WriteScopeDisplayGroup(siteCollection.Url,
                                    O14Conversions.ToScopeDisplayGroup(group, scopesIndexesForGroup, defaultScope));
                            }
                        }
                        catch (Exception ex) {
                            Logger.Log(TraceEventType.Error, ex, "Could not export a shared scope for site [{0}]", siteCollection.Url);
                        }
                        finally {
                            progress.IncreaseCounter();
                        }
                    }
                }
            }
            catch (Exception ex) {
                Logger.Log(TraceEventType.Error, ex, "Could not export the gdisplay goups for site [{0}]", siteCollection.Url);
            }
        }

        private void ExportSharedScopesInformation(SearchServiceApplicationProxy searchApp, ExportAction exportAction) {
            try {
                //get shared scopes
                List<ScopeInfo> scopesForSearchService = searchApp.GetScopesForConsumer("shared");
                if (scopesForSearchService.Count > 0) //only include in the output serachapps that have scopes
                {
                    ProgressIndicator progress =
                        new ProgressIndicator(
                            string.Format("Exporting scopes for Search App [{0}]", searchApp.Name),
                            scopesForSearchService.Count);


                    foreach (ScopeInfo scope in scopesForSearchService) {
                        try {
                            if (scope.DisplayInAdminUI) {
                                int statusCode;
                                List<RuleInfo> rulesForScope = searchApp.GetRulesInfo(scope.Id, out statusCode);
                                exportAction.WriteScope(searchApp.Name, false, O14Conversions.ToScope(scope, rulesForScope));

                                
                            }
                        }
                        catch (Exception ex) {
                            Logger.Log(TraceEventType.Error, ex, "Could not export shared scope [{0}] for Search App [{1}]", scope.Name, searchApp.Name);
                        }
                        finally {
                            progress.IncreaseCounter();
                        }
                    }

                    Schema schema = new Schema(GetSearchAppByName(searchApp.Name));

                    foreach (ManagedProperty prop in schema.AllManagedProperties) {
                        try {

                            exportAction.WriteManagedProperty(searchApp.Name, O14Conversions.ToManagedProperty(prop));
                        }
                        catch (Exception ex) {
                            Logger.Log(TraceEventType.Error, ex, "Could not export managed property [{0}] for Search App [{1}]", prop.Name, searchApp.Name);
                        }
                        finally {
                            //progress.IncreaseCounter();
                        }
                    }


                }
            }
            catch (Exception ex) {
                Logger.Log(TraceEventType.Error, ex, "Could not export shared scopes for Search App [{0}]", searchApp.Name);
            }
        }
        private void ExportManagedProperties(SearchServiceApplicationProxy searchApp, ExportAction exportAction) {
            try {

                Schema schema = new Schema(GetSearchAppByName(searchApp.Name));

                ProgressIndicator progress =
                    new ProgressIndicator(
                        string.Format("Exporting MPs for Search App [{0}]", searchApp.Name),
                        schema.AllManagedProperties.Count);



                foreach (ManagedProperty prop in schema.AllManagedProperties) {
                    try {

                        exportAction.WriteManagedProperty(searchApp.Name, O14Conversions.ToManagedProperty(prop));
                    }
                    catch (Exception ex) {
                        Logger.Log(TraceEventType.Error, ex, "Could not export managed property [{0}] for Search App [{1}]", prop.Name, searchApp.Name);
                    }
                    finally {
                        progress.IncreaseCounter();
                    }

                }
            }
            catch (Exception ex) {
                Logger.Log(TraceEventType.Error, ex, "Could not export Managed Property for Search App [{0}]", searchApp.Name);
            }
        }

        public override bool ImportManagedProperties(TManagedProperty managedPropertyElement, string entityIdentifier, ProgramOptions.ConflictBehaviorType conflictBehavior, out int createdProperties, out int updatedProperties) {
            bool shouldOverwrite = false;
            bool exists = false;
            
            PrepareToUseSearchApp(entityIdentifier);
            string entity = String.Format("Search App [{0}]", entityIdentifier);

            Schema schema = new Schema(currentSearchApp);

            var currentManagedProperties = schema.AllManagedProperties;
            

            if (currentManagedProperties.Contains(managedPropertyElement.Name)) {
                shouldOverwrite = ProgramOptions.HandleConflict(conflictBehavior, String.Format("Managed Property [{0}]", managedPropertyElement.Name));
                exists = true;
            }
            if (shouldOverwrite && exists) {
                ManagedProperty mp = currentManagedProperties[managedPropertyElement.Name];
                mp.FullTextQueriable = managedPropertyElement.FullTextQueriable;
                mp.EnabledForScoping = managedPropertyElement.EnabledForScoping;
                mp.HasMultipleValues = managedPropertyElement.HasMultipleValues;
                MappingCollection mc = new MappingCollection();
                foreach (var crawledProp in managedPropertyElement.CrawledProperty) {
                    try {
                        CrawledProperty cp = schema.GetCrawledProperty(new Guid(crawledProp.Propset), crawledProp.Name, crawledProp.VariantType);
                        mc.Add(new Mapping(cp.Propset, cp.Name, cp.VariantType, mp.PID));
                    }
                    catch (ArgumentException) {
                        Logger.Log(TraceEventType.Error,
                     "Skipped Crawled Property [{0}] in Search App [{1}] for the Managed Property [{2}] - verify name, propset or variant type",
                     crawledProp.Name, currentSearchAppName, managedPropertyElement.Name);
                    }
                }
                mp.SetMappings(mc);
                mp.Update();
                createdProperties = 0;
                updatedProperties = 1;
                return true;
            }
            else if (!exists) {
                ManagedProperty mp = currentManagedProperties.Create(managedPropertyElement.Name, O14Conversions.ToManagedPropertyType(managedPropertyElement.Type));
                mp.FullTextQueriable = managedPropertyElement.FullTextQueriable;
                mp.EnabledForScoping = managedPropertyElement.EnabledForScoping;
                mp.HasMultipleValues = managedPropertyElement.HasMultipleValues;
                MappingCollection mc = new MappingCollection();
                foreach (var crawledProp in managedPropertyElement.CrawledProperty) {
                    try {
                        CrawledProperty cp = schema.GetCrawledProperty(new Guid(crawledProp.Propset), crawledProp.Name, crawledProp.VariantType);
                        mc.Add(new Mapping(cp.Propset, cp.Name, cp.VariantType, mp.PID));
                    }
                    catch (ArgumentException) {
                        Logger.Log(TraceEventType.Error,
                     "Skipped Crawled Property [{0}] in Search App [{1}] for the Managed Property [{2}] - verify name, propset or variant type",
                     crawledProp.Name, currentSearchAppName, managedPropertyElement.Name);
                    }
                }
                mp.SetMappings(mc);
                mp.Update();
                createdProperties = 1;
                updatedProperties = 0;
                return true;
            }
            else {
                Logger.Log(TraceEventType.Error,
                     "Skipped Managed Property [{0}] in Search App [{1}] in the input file because it already existed. " +
                     "You can change the treatment of such conflicts with option conflictBehavior",
                     managedPropertyElement.Name, currentSearchAppName);
                createdProperties = 0;
                updatedProperties= 0;
                return false;
            }
        }
    }

}