﻿extern alias O12;
extern alias O14;
extern alias FS14;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Diagnostics;
using System.IO;
using System.Reflection;

using Common;

namespace SearchMigrationTool
{
    class Program
    {
        static void Main(string[] args)
        {
            Logger.FileName = Path.GetTempPath() + "SearchMigrationTool" + DateTime.UtcNow.ToFileTime().ToString() + ".log";

            ProgramOptions options = new ProgramOptions(args);

            if (!options.HasValidOptions())
            {
                Console.WriteLine(options.GetUsageString());
                return;
            }

            Console.WriteLine(options.GetProgramActionDescription());
            Console.WriteLine("Do you want to proceed?");
            if (!Util.PromptYesNo())
            {
                Console.WriteLine();
                Console.WriteLine("Aborting execution. No actions were taken.");
                return;
            }

            try
            {
                ProgramAction theAction = ProgramAction.BuildAction(options, DetectMossVersion(options));
                theAction.PerformAction();
            }
            catch (ActionException exception)
            {
                Console.WriteLine(exception.Message);
            }

            Console.WriteLine("All done.");
            if (Logger.TotalMessageCount > 0)
            {
                Console.WriteLine("Please check the logfile at [{0}] for errors and other information.", Logger.FileName);
                Logger.Close();
            }
        }

        /// <summary>
        /// Detects if a specific version of MOSS is installed by looking at the registry
        /// </summary>
        /// <param name="majorVersion"></param>
        /// <returns></returns>
        private static bool IsMossInstalled(int majorVersion)
        {
            bool isMossInstalled = false;
            string keyname = string.Format(@"SOFTWARE\Microsoft\Office Server\{0}.0",majorVersion);
            using (Microsoft.Win32.RegistryKey key = Microsoft.Win32.Registry.LocalMachine.OpenSubKey(keyname))
            {
                if (key != null)
                {
                    string version = key.GetValue("BuildVersion") as string;
                    if (version != null)
                    {
                        Version buildVersion = new Version(version);
                        if (buildVersion.Major == majorVersion)
                        {
                            isMossInstalled = true;
                        }
                    }
                }
            }
            return isMossInstalled;
        }


        /// <summary>
        /// This will build a factory, that is used for creating a layer. 
        /// Calling BuildLayer on the returned object will either create an 
        /// O12 or O14 Layer, depending on the the runtime type of the returned
        /// factory. We detect the available search assemly (O12 or O14) and return
        /// the appropriate factory.
        /// </summary>
        /// <returns></returns>
        static MossLayer DetectMossVersion(ProgramOptions options)
        {
            if (options.FastSearch)
            {
                if (IsMossInstalled(14))
                {
                    return new FS14::Layers.FS14Layer();
                }
                else
                {
                    throw new ActionException("Attempting to run against Fast Search Server for SharePoint 2010 without Microsoft Sharepoint 2010 installed.");
                }
            }
            else if (IsMossInstalled(14))
            {
                return new O14::Layers.O14Layer();
            }
            else if (IsMossInstalled(12))
            {
                return new O12::Layers.O12Layer();
            }
            else
            {
                throw new ActionException("It seems you don't have any of the Microsoft products this tool targets. It cannot continue.");
            }
        }
    }
}