using System;
using System.Collections.Specialized;
using System.IO;
using System.Text;
using System.Collections.Generic;
using System.Xml;
using Common;
using Layers;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace Tests
{
    /// <summary>
    /// Summary description for FastScopeUtilTest
    /// </summary>
    [TestClass]
    public class FastScopeUtilTest
    {
        private static TSearchObjects _searchObjectsElement;
        private const string TestDataFileName = "TestData.xml";

        [ClassInitialize()]
        public static void FastScopeUtilTestInitialize(TestContext testContext)
        {
            Logger.FileName = Path.GetTempPath() + "SearchMigration" +
                              DateTime.UtcNow.ToFileTime() + ".log";
            _searchObjectsElement = DeserializeXml(GetResourceString(TestDataFileName));
        }

        [ClassCleanup()]
        public static void FastScopeUtilTestCleanup()
        {
            Console.WriteLine("Please check the logfile at [{0}] for errors and other information.", Logger.FileName);
            Logger.Close();
        }

        /// <summary>
        ///Gets or sets the test context which provides
        ///information about and functionality for the current test run.
        ///</summary>
        public TestContext TestContext { get; set; }

        #region Additional test attributes

        //
        // You can use the following additional attributes as you write your tests:
        //
        // Use ClassInitialize to run code before running the first test in the class
        // [ClassInitialize()]
        // public static void MyClassInitialize(TestContext testContext) { }
        //
        // Use ClassCleanup to run code after all tests in a class have run
        // [ClassCleanup()]
        // public static void MyClassCleanup() { }
        //
        // Use TestInitialize to run code before running each test 
        // [TestInitialize()]
        // public void MyTestInitialize() { }
        //
        // Use TestCleanup to run code after each test has run
        // [TestCleanup()]
        // public void MyTestCleanup() { }
        //

        #endregion

        [TestMethod]
        public void NegativeTests()
        {
            const string ssaName = "Search Service Application";
            const string scopeName = "negativescope";

            // null-valued scopeElement
            string filter = FastScopeUtil.CreateFastScopeFilter(null, null, ssaName);
            Assert.AreEqual("", filter);

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            // null-valued ManagedProperty list
            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"not(contentclass:""urn:content-class:spspeople"")", filter);

            // empty ManagedProperty list
            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, new StringCollection(), searchApp.Name);
            Assert.AreEqual(@"not(contentclass:""urn:content-class:spspeople"")", filter);
        }

        [TestMethod]
        public void OnlyFilterRule()
        {
            const string ssaName = "Search Service Application";
            const string scopeName = "onlyfilter";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);

            // Filter is preserved
            Assert.AreEqual(@"not(contentclass:spspeople)", filter);
        }

        [TestMethod]
        public void TestIncludePropertyScopes()
        {
            const string ssaName = "Search Service Application";
            string scopeName = "includeproperty";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"contentclass:""i1""", filter);

            scopeName = "includeproperty2";

            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"or(contentclass:""i1"", contentclass:""i2"")", filter);

            // contentclass not in ManagedProperty list
            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, new StringCollection() { "nothere" }, searchApp.Name);
            Assert.AreEqual(@"", filter);
        }

        [TestMethod]
        public void TestRequirePropertyScopes()
        {
            const string ssaName = "Search Service Application";
            string scopeName = "requireproperty";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"contentclass:""r1""", filter);

            scopeName = "requireproperty2";

            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"and(contentclass:""r1"", contentclass:""r2"")", filter);

            // contentclass not in ManagedProperty list
            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, new StringCollection() { "nothere" }, searchApp.Name);
            Assert.AreEqual(@"", filter);
        }

        [TestMethod]
        public void TestExcludePropertyScopes()
        {
            const string ssaName = "Search Service Application";
            string scopeName = "excludeproperty";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"not(contentclass:""e1"")", filter);

            scopeName = "excludeproperty2";

            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"or(not(contentclass:""e1""), not(contentclass:""e2""))", filter);

            // contentclass not in ManagedProperty list
            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, new StringCollection() { "nothere" }, searchApp.Name);
            Assert.AreEqual(@"", filter);
        }

        [TestMethod]
        public void TestIncludeRequirePropertyScopes()
        {
            const string ssaName = "Search Service Application";
            string scopeName = "includerequireproperty";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"and(contentclass:""i1"", contentclass:""r1"")", filter);

            scopeName = "includerequireproperty2";

            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"and(or(contentclass:""i1"", contentclass:""i2""), contentclass:""r1"")", filter);

            scopeName = "includerequireproperty3";

            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"and(contentclass:""i1"", and(contentclass:""r1"", contentclass:""r2""))", filter);
        }

        [TestMethod]
        public void TestIncludeExcludePropertyScopes()
        {
            const string ssaName = "Search Service Application";
            string scopeName = "includeexcludeproperty";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"andnot(contentclass:""i1"", contentclass:""e1"")", filter);

            scopeName = "includeexcludeproperty2";

            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"andnot(path:starts-with(""http://i1""), path:starts-with(""http://e1""), path:starts-with(""http://e2""))", filter);

            scopeName = "includeexcludeproperty3";

            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"andnot(or(path:starts-with(""http://i1""), path:starts-with(""http://i2"")), path:starts-with(""http://e1""), path:starts-with(""http://e2""))", filter);
        }

        [TestMethod]
        public void TestRequireExcludePropertyScopes()
        {
            const string ssaName = "Search Service Application";
            const string scopeName = "requireexcludeproperty";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"andnot(contentclass:""r1"", contentclass:""e1"")", filter);
        }

        [TestMethod]
        public void TestIncludeRequireExcludePropertyScopes()
        {
            const string ssaName = "Search Service Application";
            const string scopeName = "includerequireexcludeproperty";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"andnot(and(contentclass:""i1"", contentclass:""r1""), contentclass:""e1"")", filter);
        }

        [TestMethod]
        public void TestIncludeUrlScopes()
        {
            const string ssaName = "Search Service Application";
            string scopeName = "includefolder";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"path:starts-with(""http://i1"")", filter);

            scopeName = "includefolder2";

            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"or(path:starts-with(""http://i1""), path:starts-with(""http://i2""))", filter);

            // path not in ManagedProperty list
            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, new StringCollection() { "nothere" }, searchApp.Name);
            Assert.AreEqual(@"", filter);
        }

        [TestMethod]
        public void TestRequireUrlScopes()
        {
            const string ssaName = "Search Service Application";
            string scopeName = "requirefolder";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"path:starts-with(""http://r1"")", filter);

            scopeName = "requirefolder2";
            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"and(path:starts-with(""http://r1""), path:starts-with(""http://r2""))", filter);

            // path not in ManagedProperty list
            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, new StringCollection() { "nothere" }, searchApp.Name);
            Assert.AreEqual(@"", filter);
        }

        [TestMethod]
        public void TestExcludeUrlScopes()
        {
            const string ssaName = "Search Service Application";
            string scopeName = "excludefolder";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"not(path:starts-with(""http://e1""))", filter);

            scopeName = "excludefolder2";
            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"or(not(path:starts-with(""http://e1"")), not(path:starts-with(""http://e2"")))", filter);

            // path not in ManagedProperty list
            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, new StringCollection() { "nothere" }, searchApp.Name);
            Assert.AreEqual(@"", filter);
        }

        [TestMethod]
        public void TestIncludeRequireUrlScopes()
        {
            const string ssaName = "Search Service Application";
            string scopeName = "includerequirefolder";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"and(path:starts-with(""http://i1""), path:starts-with(""http://r1""))", filter);

            scopeName = "includerequirefolder2";
            scopeElement = FetchScope(scopeName, searchApp.Scope);

            filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"and(or(path:starts-with(""http://i1""), path:starts-with(""http://i2"")), path:starts-with(""http://r1""))", filter);
        }

        [TestMethod]
        public void TestIncludeDomain()
        {
            const string ssaName = "Search Service Application";
            const string scopeName = "includedomain";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"domain:""domain.com""", filter);
        }

        [TestMethod]
        public void TestVeryLongFilter()
        {
            const string ssaName = "Search Service Application";
            const string scopeName = "verylongfilter";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"", filter);
        }

        [TestMethod]
        public void TestIncludeHostName()
        {
            const string ssaName = "Search Service Application";
            const string scopeName = "includehostname";

            TSearchApplication searchApp = FetchSearchApp(ssaName, _searchObjectsElement.SearchApplication);
            TScope scopeElement = FetchScope(scopeName, searchApp.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, searchApp.Name);
            Assert.AreEqual(@"domain:starts-with(""hostname"")", filter);
        }

        [TestMethod]
        public void TestSiteCollectionScope()
        {
            const string siteCollectionUrl = "http://sitecollection/";
            const string scopeName = "sitecollectionscope";

            TSiteCollection siteColl = FetchSiteCollection(siteCollectionUrl, _searchObjectsElement.SiteCollection);
            TScope scopeElement = FetchScope(scopeName, siteColl.Scope);

            string filter = FastScopeUtil.CreateFastScopeFilter(scopeElement, null, siteColl.Url);
            Assert.AreEqual(@"path:starts-with(""http://i1"")", filter);
        }

        [TestMethod]
        public void TestFilterBuilders()
        {
            Assert.AreEqual("", FastScopeUtil.BuildGenericFqlFilter("and", null));
            Assert.AreEqual("", FastScopeUtil.BuildGenericFqlFilter("and", new List<string> { null }));
            Assert.AreEqual("", FastScopeUtil.BuildGenericFqlFilter("", new List<string> { null }));
            Assert.AreEqual("term1", FastScopeUtil.BuildGenericFqlFilter("", new List<string> { "term1" }));
            Assert.AreEqual("term1", FastScopeUtil.BuildGenericFqlFilter("and", new List<string> { "term1" }));
            Assert.AreEqual("and(term1, term2)",
                            FastScopeUtil.BuildGenericFqlFilter("and", new List<string> { "term1", "term2" }));

            Assert.AreEqual("term1", FastScopeUtil.BuildFqlAndFilter(new List<string> { "term1" }));
            Assert.AreEqual("and(term1, term2)", FastScopeUtil.BuildFqlAndFilter(new List<string> { "term1", "term2" }));

            Assert.AreEqual("term1", FastScopeUtil.BuildFqlOrFilter(new List<string> { "term1" }));
            Assert.AreEqual("or(term1, term2)", FastScopeUtil.BuildFqlOrFilter(new List<string> { "term1", "term2" }));

            Assert.AreEqual("not(term1)", FastScopeUtil.BuildFqlNotFilter(new List<string> { "term1" }));
            Assert.AreEqual("not(term1, term2)", FastScopeUtil.BuildFqlNotFilter(new List<string> { "term1", "term2" }));

            Assert.AreEqual("andnot(term1, term2)",
                            FastScopeUtil.BuildFqlAndNotFilter(new List<string> { "term1", "term2" }));
            Assert.AreEqual("andnot(term1, term2, term3)",
                            FastScopeUtil.BuildFqlAndNotFilter(new List<string> { "term1", "term2", "term3" }));

            Assert.AreEqual("and(term1, term2)", FastScopeUtil.CombineFilters("and", "term1", "term2"));
        }

        private static TScope FetchScope(string scopeName, TScope[] scopes)
        {
            foreach (var scope in scopes)
            {
                if (scope.Name.Equals(scopeName))
                    return scope;
            }
            throw new ArgumentException(String.Format("Scope {0} not found.", scopeName), scopeName);
        }

        private static TSearchApplication FetchSearchApp(string searchAppName, TSearchApplication[] applications)
        {
            foreach (var application in applications)
            {
                if (application.Name.Equals(searchAppName))
                    return application;
            }
            throw new ArgumentException(String.Format("Search App {0} not found", searchAppName), searchAppName);
        }

        private static TSiteCollection FetchSiteCollection(string siteCollectionUrl, TSiteCollection[] siteCollections)
        {
            foreach (TSiteCollection siteCollectionElement in siteCollections)
            {
                if (siteCollectionElement.Url.Equals(siteCollectionUrl))
                    return siteCollectionElement;
            }
            throw new ArgumentException(String.Format("Site Collection with URL {0} not found", siteCollectionUrl), siteCollectionUrl);
        }

        public static Stream GetResourceStream(string resourceId)
        {
            System.Reflection.Assembly assembly = System.Reflection.Assembly.GetAssembly(typeof(FastScopeUtilTest));
            if (assembly != null)
            {
                return assembly.GetManifestResourceStream(typeof(FastScopeUtilTest), resourceId);
            }
            return null;
        }

        public static string GetResourceString(string resourceId)
        {

            Stream stream = GetResourceStream(resourceId);
            if (null != stream)
            {
                return (new StreamReader(stream)).ReadToEnd();
            }
            return null;
        }

        private static TSearchObjects DeserializeXml(string xml)
        {
            byte[] byteArray = Encoding.ASCII.GetBytes(xml);
            MemoryStream stream = new MemoryStream(byteArray);

            System.Xml.Serialization.XmlSerializer serializer =
                new System.Xml.Serialization.XmlSerializer(typeof(TSearchObjects));
            XmlReader reader = new XmlTextReader(stream);

            if (serializer.CanDeserialize(reader))
            {
                return (TSearchObjects)serializer.Deserialize(reader);
            }
            throw new ArgumentException("Unable to deserialize XML", xml);
        }
    }
}
